/******************************************************************************

    $Id: alloc.c,v 1.3 2003/08/28 20:51:57 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/alloc.c,v $

    Wrappers for malloc() and friends.

    cpp_malloc(), cpp_calloc(), cpp_realloc(), and cpp_free() is simple
    wrappers for the ones provided by the system. They are used to keep
    track of the number of allocated blocks in 'CPP.blocks_allocated'.
    This counter is used to check that all memory has endeed been
    released when the library is done.

    cpp_safe_malloc(), cpp_safe_calloc(), cpp_safe_realloc() are
    similar to their unsafe versions, except that the result is
    checked, and it is guaranteed that they have not failed if they
    return. In case of allocation error, it is considered a fatal error,
    and control is passed to the main library entry point in "cpp.c"
    using longjmp().

******************************************************************************/

#include "priv.h"


void* cpp_malloc(CPP* cpp, size_t size)
{
    void* ptr = malloc(size);

    if ( ptr ) {
        ++(cpp->alloc_count);
    }

    return ptr;
}


void* cpp_calloc(CPP* cpp, size_t num, size_t size)
{
    void* ptr = calloc(num, size);

    if ( ptr ) {
        ++(cpp->alloc_count);
    }

    return ptr;
}


void* cpp_realloc(CPP* cpp, void* ptr, size_t size)
{
    void* new_ptr = realloc(ptr, size);

    if ( ptr ) {
        ++(cpp->dealloc_count);
    }
    if ( new_ptr ) {
        ++(cpp->alloc_count);
    }

    return new_ptr;
}


void cpp_free(CPP* cpp, void* ptr)
{
    if ( ptr ) {
        free(ptr);
        ++(cpp->dealloc_count);
    }
}


void cpp_out_of_memory(CPP* cpp)
{
    cpp_fatal(cpp, "Out of memory");
}


void* cpp_safe_malloc(CPP* cpp, size_t size)
{
    void* ptr = cpp_malloc(cpp, size);

    if ( ptr==NULL && size ) {
        cpp_out_of_memory(cpp);
    }

    return ptr;
}


void* cpp_safe_calloc(CPP* cpp, size_t num, size_t size)
{
    void* ptr = cpp_calloc(cpp, num, size);

    if ( ptr==NULL && num && size ) {
        cpp_out_of_memory(cpp);
    }

    return ptr;
}


void* cpp_safe_realloc(CPP* cpp, void* ptr, size_t size)
{
    ptr = cpp_realloc(cpp, ptr, size);

    if ( ptr==NULL && size ) {
        cpp_out_of_memory(cpp);
    }

    return ptr;
}
