/******************************************************************************

    $Id: cond.c,v 1.3 2003/09/05 20:30:27 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/cond.c,v $

******************************************************************************/

#include "priv.h"


static CONDITIONAL* cpp_stack_conditional(CPP* cpp)
{
    INPUT*          input = cpp->input;
    CONDITIONAL*    cond;

    if ( input->cond_top == cpp->cond_stack_size ) {
        size_t n = cpp->cond_stack_size * 2;
        cpp->cond_stack = (CONDITIONAL*)cpp_safe_realloc(cpp, cpp->cond_stack, n * sizeof(CONDITIONAL));
        cpp->cond_stack_size = n;
    }

    cond = cpp->cond_stack + input->cond_top;
    input->cond_top++;

    memset(cond, 0, sizeof(CONDITIONAL));

    return cond;
}


static CONDITIONAL* cpp_get_conditional(CPP* cpp)
{
    INPUT* input = cpp->input;

    return input->cond_top 
        ? cpp->cond_stack + input->cond_top - 1
        : NULL;
}


bool cpp_is_conditional_true(CPP* cpp)
{
    CONDITIONAL* cond = cpp_get_conditional(cpp);

    return cond ? cond->is_true : true;
}


void cpp_parse_if(CPP* cpp, const char* pos)
{
    CONDITIONAL*    cond;
    bool            is_true;

    is_true = cpp_is_conditional_true(cpp);
    cond = cpp_stack_conditional(cpp);
    cond->file = cpp->file;
    cond->line = cpp->line;
    cond->ident = "ifdef";
    if ( is_true ) {
        is_true = cpp_eval_if_expression(cpp, "if", pos);
        if ( is_true ) {
            cond->is_true = true;
            cond->skip_rest = true;
        } else {
            cond->is_true = false;
            cond->skip_rest = false;
        }
    } else {
        cond->is_true = false;
        cond->skip_rest = true;
    }
}


static void cpp_parse_ifdef_ifndef(CPP* cpp, const char* pos, bool is_ifdef)
{
    CONDITIONAL*    cond;
    PPTOKEN*        token;
    int             tag;
    size_t          len;
    const char*     name;
    bool            is_true;
    const char*     ident = is_ifdef ? "ifdef" : "ifndef";

    // Get identifier
    pos = cpp_skipspace(pos);
    cpp_identify_token(pos, &tag, &len);
    if ( tag != T_IDENT ) {
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "Macro name expected after '#");
        cpp_msg_cats(cpp, ident);
        cpp_msg_cats(cpp, "'");
        if ( tag != T_END ) {
            cpp_msg_cats(cpp, ", found ");
            cpp_msg_cats(cpp, cpp_strtag(tag));
        }
        token = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
        cpp_output_append(cpp, token);
        return;
    }
    name = cpp_heap_strndup(&cpp->output_heap, pos, len);
    pos = cpp_skipspace(pos+len);

    // Check for end-of-line
    pos = cpp_skipspace(pos);
    if ( *pos ) {
        cpp_identify_token(pos, &tag, &len);
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "Token unexpected after '#");
        cpp_msg_cats(cpp, ident);
        cpp_msg_cats(cpp, " ");
        cpp_msg_cats(cpp, name);
        cpp_msg_cats(cpp, "': ");
        cpp_msg_cats(cpp, cpp_strtag(tag));
        token = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
        cpp_output_append(cpp, token);
    }

    is_true = cpp_is_conditional_true(cpp);
    cond = cpp_stack_conditional(cpp);
    cond->file = cpp->file;
    cond->line = cpp->line;
    cond->ident = "ifdef";
    if ( is_true ) {
        MACRO* macro = cpp_find_macro(cpp, name);
        if ( macro ) {
            is_true = is_ifdef;
        } else {
            is_true = !is_ifdef;
        }
        if ( is_true ) {
            cond->is_true = true;
            cond->skip_rest = true;
        } else {
            cond->is_true = false;
            cond->skip_rest = false;
        }
    } else {
        cond->is_true = false;
        cond->skip_rest = true;
    }
}


void cpp_parse_ifdef(CPP* cpp, const char* pos)
{
    cpp_parse_ifdef_ifndef(cpp, pos, true);
}


void cpp_parse_ifndef(CPP* cpp, const char* pos)
{
    cpp_parse_ifdef_ifndef(cpp, pos, false);
}


void cpp_parse_elif(CPP* cpp, const char* pos)
{
    CONDITIONAL*    cond;
    INPUT*          input;
    PPTOKEN*        token;

    input = cpp->input;
    if ( input->cond_top == 0 
      || (input->parent && input->cond_top == input->parent->cond_top)
       )
    {
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "'#elif' without '#if', '#ifdef', or '#ifndef'");
        token = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
        cpp_output_append(cpp, token);
        return;
    }

    cond = cpp_get_conditional(cpp);
    if ( cond->else_file ) {
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "'#elif' cannot follow '#else' found at ");
        cpp_msg_catfl(cpp, cond->else_file, cond->else_line);
        token = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
        cpp_output_append(cpp, token);
        return;
    }

    if ( cond->skip_rest || cond->is_true ) {
        cond->is_true = false;
    } else {
        cond->is_true = cpp_eval_if_expression(cpp, "elif", pos);
        cond->skip_rest = !cond->is_true;
    }
}


void cpp_parse_else(CPP* cpp, const char* pos)
{
    CONDITIONAL*    cond;
    INPUT*          input;
    PPTOKEN*        token;
    int             tag;
    size_t          len;

    // Check for end-of-line
    pos = cpp_skipspace(pos);
    if ( *pos ) {
        cpp_identify_token(pos, &tag, &len);
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "Token unexpected after '#else': ");
        cpp_msg_cats(cpp, cpp_strtag(tag));
        token = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
        cpp_output_append(cpp, token);
    }

    // Check if there is an active conditional block
    input = cpp->input;
    if ( input->cond_top == 0 
      || (input->parent && input->cond_top == input->parent->cond_top)
       )
    {
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "'#else' without '#if', '#ifdef', or '#ifndef'");
        token = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
        cpp_output_append(cpp, token);
        return;
    }

    cond = cpp_get_conditional(cpp);
    if ( cond->else_file ) {
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "'#else' already found at ");
        cpp_msg_catfl(cpp, cond->else_file, cond->else_line);
        token = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
        cpp_output_append(cpp, token);
        return;
    }

    if ( cond->skip_rest ) {
        cond->is_true = false;
    } else {
        cond->is_true = true;
        cond->skip_rest = true;
    }
}


void cpp_parse_endif(CPP* cpp, const char* pos)
{
    INPUT*      input;
    PPTOKEN*    token;
    int         tag;
    size_t      len;

    // Check for end-of-line
    pos = cpp_skipspace(pos);
    if ( *pos ) {
        cpp_identify_token(pos, &tag, &len);
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "Token unexpected after '#endif': ");
        cpp_msg_cats(cpp, cpp_strtag(tag));
        token = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
        cpp_output_append(cpp, token);
    }

    // Check if there is an active conditional block
    input = cpp->input;
    if ( input->cond_top == 0 
      || (input->parent && input->cond_top == input->parent->cond_top)
       )
    {
        cpp_identify_token(pos,&tag,&len);
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "'#endif' without '#if', '#ifdef', or '#ifndef'");
        cpp_msg_cats(cpp, cpp_strtag(tag));
        token = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
        cpp_output_append(cpp, token);
        return;
    }

    input->cond_top--;
}

