/******************************************************************************

    $Id: cpp.c,v 1.6 2003/08/28 19:12:27 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/cpp.c,v $

    This module contains the public interface to the library.
    
******************************************************************************/

#include "priv.h"

/******************************************************************************
    Initialization
******************************************************************************/

CPP* cpp_new(void)
{
    CPP* cpp;

    // Allocate main context
    cpp = (CPP*)calloc(1, sizeof(CPP));
    if ( cpp == NULL ) {
        return NULL;
    }

    // Exception handler
    if ( setjmp(cpp->fatal_exit) ) {
        cpp_delete(cpp);
        return NULL;
    }

    cpp_heap_init(cpp, &cpp->cpp_heap);
    cpp_heap_init(cpp, &cpp->unit_heap);
    cpp_heap_init(cpp, &cpp->output_heap);
    cpp->macro_hsize = 2047;
    cpp->macro_htab = (MACRO**)cpp_safe_calloc(cpp, cpp->macro_hsize, sizeof(MACRO*));
    cpp_buffer_init(cpp, &cpp->input_buf);
    cpp_buffer_init(cpp, &cpp->msg_buf);
    cpp_buffer_init(cpp, &cpp->scratch);
    cpp->input_ptr = "";
    cpp->cond_stack_size = 16;
    cpp->cond_stack = (CONDITIONAL*)cpp_safe_calloc(cpp, cpp->cond_stack_size, sizeof(CONDITIONAL));
    cpp->expand_pragma = true;

    return cpp;
}

/******************************************************************************
    Destruction
******************************************************************************/

void cpp_delete(CPP* cpp)
{
    if (cpp) {
        cpp_input_popall(cpp);
        cpp_buffer_cleanup(&cpp->scratch);
        cpp_buffer_cleanup(&cpp->input_buf);
        cpp_buffer_cleanup(&cpp->msg_buf);
        cpp_free(cpp, cpp->macro_htab);
        cpp_heap_clear(&cpp->output_heap);
        cpp_heap_clear(&cpp->unit_heap);
        cpp_heap_clear(&cpp->cpp_heap);
        cpp_heap_cleanup(cpp);
        cpp_free(cpp, cpp->cond_stack);
        assert(cpp->alloc_count == cpp->dealloc_count);
        free(cpp);
    }
}

/******************************************************************************
    Error handling
******************************************************************************/

void cpp_fatal(CPP* cpp, const char* msg)
{
    cpp->fatal_msg = msg;
    longjmp(cpp->fatal_exit, -1);
}

/******************************************************************************
    Configuration
******************************************************************************/

const char* cpp_define(CPP* cpp, const char* assignment)
{
    const char* eq;

    if ( setjmp(cpp->fatal_exit) ) {
        return cpp->fatal_msg;
    }

    eq = strchr(assignment, '=');
    if ( eq ) {
        cpp_config_define(cpp, assignment, eq-assignment, eq+1);
    } else {
        cpp_config_define(cpp, assignment, strlen(assignment), "1");
    }

    return NULL;
}


const char* cpp_undefine(CPP* cpp, const char* name)
{
    if ( setjmp(cpp->fatal_exit) ) {
        return cpp->fatal_msg;
    }

    cpp_config_define(cpp, name, strlen(name), NULL);

    return NULL;
}


const char* cpp_include(CPP* cpp, const char* path)
{
    if ( setjmp(cpp->fatal_exit) ) {
        return cpp->fatal_msg;
    }

    cpp_config_include(cpp, path);

    return NULL;
}


const char* cpp_include_list(CPP* cpp, const char* list)
{
    if ( setjmp(cpp->fatal_exit) ) {
        return cpp->fatal_msg;
    }

    cpp_config_include_list(cpp, list);

    return NULL;
}


/******************************************************************************
    Preprocessing
******************************************************************************/

const char* cpp_open(CPP* cpp, const char* path, FILE* fp)
{
    // Exception handler
    if ( setjmp(cpp->fatal_exit) ) {
        return cpp->fatal_msg;
    }

    // Reset everything
    cpp_input_popall(cpp);
    cpp_heap_clear(&cpp->unit_heap);
    cpp_heap_clear(&cpp->output_heap);
    cpp_buffer_clear(&cpp->input_buf);
    cpp->input_ptr = "";
    cpp->input_freelist = NULL;
    cpp_list_reset(&cpp->output_queue);
    cpp->output_state = 0;
    memset(cpp->macro_htab, 0, cpp->macro_hsize * sizeof(MACRO*));
    cpp->macro_count = 0;
    cpp_macro_prepare(cpp);
    cpp->push_back = NULL;
    cpp->is_including = true;

    // Initialize new input
    cpp_input_push(cpp, path);
    if ( fp == NULL ) {
        if ( (fp = fopen(path, "r")) == NULL ) {
            cpp_fatal(cpp, strerror(errno));
        }
        cpp->input->owner = true;
    }
    cpp->input->fp = fp;

    return NULL;
}


const char* cpp_gettoken(CPP* cpp, TOKEN* token)
{
    PPTOKEN* pptoken;

    // See if token is ready on output queue
    pptoken = cpp_output_get(cpp);
    if ( pptoken ) {
        *token = pptoken->token;
        return NULL;
    }

    // Exception handler
    if ( setjmp(cpp->fatal_exit) ) {
        return cpp->fatal_msg;
    }

    // Clean up after previously returned tokens
    cpp_heap_clear(&cpp->output_heap);

    // Parse some input, producing output
    do {
        cpp_parse(cpp);
    } while ( cpp->output_queue.first==NULL || cpp->output_state );
    cpp_listfix(cpp, &cpp->output_queue);
    pptoken = cpp_output_get(cpp);

    *token = pptoken->token;

    return NULL;
}
