/******************************************************************************

    $Id: entok.c,v 1.2 2003/08/28 21:59:29 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/entok.c,v $

******************************************************************************/

#include "priv.h"

// Determine length of header_name
size_t cpp_header_name_len(const char* pos)
{
    const char* end;

    switch ( *pos ) {
        case '"':
            end = strchr(pos+1, '"');
            break;
        case '<':
            end = strchr(pos+1, '>');
            break;
        default:
            return 0;
    }

    return (end && end > pos+1) ? (end-pos+1) : 0;
}


// Determine length of name in #line directive
size_t cpp_line_name_len(const char* pos)
{
    if ( *pos == '"' ) {
        const char* end = strchr(pos+1, '"');
        return end ? (end-pos+1) : 0;
    } else {
        return 0;
    }
}


// Determine length of digit sequence
size_t cpp_digit_sequence_len(const char* pos)
{
    size_t len = 0;

    while ( ISDIGIT(pos[len]) ) {
        ++len;
    }

    return len;
}


// Determine length of a universal_character_name
size_t cpp_universal_char_len(const char* pos)
{
    int i;

    if ( pos[0]=='\\' ) {
        if ( pos[1]=='u' ) {
            pos += 2;
            for ( i = 4 ; i ; i-- ) {
                if ( !ISXDIGIT(*pos) ) {
                    return 0;
                }
                ++pos;
            }
            return 6;
        } else if ( pos[1]=='U' ) {
            pos += 2;
            for ( i = 8 ; i ; i-- ) {
                if ( !ISXDIGIT(*pos) ) {
                    return 0;
                }
                ++pos;
            }
            return 10;
        }
    }

    return 0;
}


// Determine length of identifier
size_t cpp_identifier_len(const char* pos)
{
    size_t n;
    size_t len = cpp_universal_char_len(pos);

    if ( len || (!ISDIGIT(*pos) && ISIDENT(*pos)) ) {
        for (;;) {
            if ( ISIDENT(pos[len]) ) {
                ++len;
            } else if ( (n=cpp_universal_char_len(pos+len)) != 0 ) {
                len += n;
            } else {
                break;
            }
        }
    }

    return len;
}


// Determine length of character or string literal
size_t cpp_text_constant_len(const char* pos)
{
    size_t len = 0;

    if ( (pos[0]=='\'' || pos[0]=='"') 
      || (pos[0]=='L' && (pos[1]=='\'' || pos[1]=='\"')) 
       ) 
    {
        char quote = *pos;
        if (quote=='L') {
            len = 2;
            quote = pos[1];
        } else {
            len = 1;
        }
        while ( pos[len] != quote ) {
            if ( pos[len] == '\\' ) {
                ++len;
            }
            ++len;
        }
        ++len;
    }

    return len;
}


// Determine length of preprocessing number
size_t cpp_pp_number_len(const char* pos)
{
    size_t len;

    if ( ISDIGIT(pos[0]) ) {
        len = 1;
    } else if (pos[0]=='.' && ISDIGIT(pos[1])) {
        len = 2;
    } else {
        return 0;
    }

    for (;;) {
        if ( ISIDENT(pos[len]) ) {
            ++len;
        } else if ( (pos[len]=='e' || pos[len]=='E') 
                 && (pos[len+1]=='+' || pos[len+1]=='-') 
                  ) 
        {
            ++len;
            ++len;
        } else {
            return len;
        }
    }
}


// Identify token, returning tag and length
void cpp_identify_token(const char* pos, int* tag, size_t* len)
{
    char        first;
    const char* end;

    first = *pos;

    switch (first) {
        case '\0':
            *len = 0;
            *tag = T_END;
            return;
        case ' ':
        case '\f':
        case '\n':
        case '\r':
        case '\t':
        case '\v':
            end = pos+1;
            while ( ISSPACE(*end) ) {
                ++end;
            }
            *len = end - pos;
            *tag = T_SPACE;
            return;
        case '\"':
        case 'L':
        case '\'':
            *len = cpp_text_constant_len(pos);
            if (*len) {
                char quote = (first=='L') ? pos[1] : first;
                *tag = quote=='"' ? T_STRLIT : T_CHRLIT;
                return;
            }
            break;
        case '/':
            switch (pos[1]) {
                case '=':
                    *len = 2;
                    *tag = T_DIV_ASSIGN;
                    return;
                default:
                    *len = 1;
                    *tag = T_DIV;
                    return;
            }
        case '*':
            switch (pos[1]) {
                case '=':
                    *len = 2;
                    *tag = T_MUL_ASSIGN;
                    return;
                default:
                    *len = 1;
                    *tag = T_MUL;
                    return;
            }
        case '+':
            switch (pos[1]) {
                case '+':
                    *len = 2;
                    *tag = T_INC;
                    return;
                case '=':
                    *len = 2;
                    *tag = T_ADD_ASSIGN;
                    return;
                default:
                    *len = 1;
                    *tag = T_ADD;
                    return;
            }
        case '-':
            switch (pos[1]) {
                case '>':
                    *len = 2;
                    *tag = T_ARROW;
                    return;
                case '-':
                    *len = 2;
                    *tag = T_DEC;
                    return;
                case '=':
                    *len = 2;
                    *tag = T_SUB_ASSIGN;
                    return;
                default:
                    *len = 1;
                    *tag = T_SUB;
                    return;
            }
        case '&':
            switch (pos[1]) {
                case '&':
                    *len = 2;
                    *tag = T_AND_AND;
                    return;
                case '=':
                    *len = 2;
                    *tag = T_AND_ASSIGN;
                    return;
                default:
                    *len = 1;
                    *tag = T_AND;
                    return;
            }
        case '|':
            switch (pos[1]) {
                case '|':
                    *len = 2;
                    *tag = T_OR_OR;
                    return;
                case '=':
                    *len = 2;
                    *tag = T_OR_ASSIGN;
                    return;
                default:
                    *len = 1;
                    *tag = T_OR;
                    return;
            }
        case '%':
            switch (pos[1]) {
                case ':':
                    if (pos[2]=='%' && pos[3]==':') {
                        *len = 4;
                        *tag = T_HASH_HASH;
                    } else {
                        *len = 2;
                        *tag = T_HASH;
                    }
                    return;
                case '>':
                    *len = 2;
                    *tag = T_CURLY_CLOSE;
                    return;
                case '=':
                    *len = 2;
                    *tag = T_MOD_ASSIGN;
                    return;
                default:
                    *len = 1;
                    *tag = T_MOD;
                    return;
            }
        case '^':
            switch (pos[1]) {
                case '=':
                    *len = 2;
                    *tag = T_XOR_ASSIGN;
                    return;
                default:
                    *len = 1;
                    *tag = T_XOR;
                    return;
            }
        case '!':
            switch (pos[1]) {
                case '=':
                    *len = 2;
                    *tag = T_NE;
                    return;
                default:
                    *len = 1;
                    *tag = T_NOT;
                    return;
            }
        case '#':
            switch (pos[1]) {
                case '#':
                    *len = 2;
                    *tag = T_HASH_HASH;
                    return;
                default:
                    *len = 1;
                    *tag = T_HASH;
                    return;
            }
        case '<':
            switch (pos[1]) {
                case '<':
                    if ( pos[2]=='=' ) {
                        *len = 3;
                        *tag = T_SHL_ASSIGN;
                    } else {
                        *len = 2;
                        *tag = T_SHL;
                    }
                    return;
                case '=':
                    *len = 2;
                    *tag = T_LE;
                    return;
                case ':':
                    *len = 2;
                    *tag = T_SHARP_OPEN;
                    return;
                case '%':
                    *len = 2;
                    *tag = T_CURLY_OPEN;
                    return;
                default:
                    *len = 1;
                    *tag = T_LT;
                    return;
            }
        case '>':
            switch (pos[1]) {
                case '>':
                    if (pos[2]=='=') {
                        *len = 3;
                        *tag = T_SHR_ASSIGN;
                    } else {
                        *len =  2;
                        *tag = T_SHR;
                    }
                    return;
                case '=':
                    *len = 2;
                    *tag = T_GE;
                    return;
                default:
                    *len = 1;
                    *tag = T_GT;
                    return;
            }
        case '=':
            switch (pos[1]) {
                case '=':
                    *len = 2;
                    *tag = T_EQ;
                    return;
                default:
                    *len = 1;
                    *tag = T_ASSIGN;
                    return;
            }
        case '.':
            switch (pos[1]) {
                case '.':
                    if (pos[2]=='.') {
                        *len = 3;
                        *tag = T_ELLIPSIS;
                    } else {
                        *len = 1;
                        *tag = T_DOT;
                    }
                    return;
                default:
                    *len = cpp_pp_number_len(pos);
                    if ( *len ) {
                        *tag = T_NUMLIT;
                    } else {
                        *len = 1;
                        *tag = T_DOT;
                    }
                    return;
            }
        case ':':
            switch (pos[1]) {
                case '>':
                    *len = 2;
                    *tag = T_SHARP_CLOSE;
                    return;
                default:
                    *len = 1;
                    *tag = T_COLON;
                    return;
            }
        case '[':
            *len = 1;
            *tag = T_SHARP_OPEN;
            return;
        case ']':
            *len = 1;
            *tag = T_SHARP_CLOSE;
            return;
        case '{':
            *len = 1;
            *tag = T_CURLY_OPEN;
            return;
        case '}':
            *len = 1;
            *tag = T_CURLY_CLOSE;
            return;
        case '(':
            *len = 1;
            *tag = T_PAREN_OPEN;
            return;
        case ')':
            *len = 1;
            *tag = T_PAREN_CLOSE;
            return;
        case ',':
            *len = 1;
            *tag = T_COMMA;
            return;
        case ';':
            *len = 1;
            *tag = T_SEMI;
            return;
        case '?':
            *len = 1;
            *tag = T_QMARK;
            return;
        case '~':
            *len = 1;
            *tag = T_NEG;
            return;
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
            *len = cpp_pp_number_len(pos);
            *tag = T_NUMLIT;
            return;
    }

    // Identifiers
    *len = cpp_identifier_len(pos);
    if ( *len ) {
        *tag = T_IDENT;
        return;
    }

    // Unrecognized character
    *len = 1;
    *tag = T_BAD;
}
