/******************************************************************************

    $Id: expr.c,v 1.2 2003/08/28 21:59:29 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/expr.c,v $

******************************************************************************/

#include "priv.h"
#include "expr.h"


// unary minus operator applied to unsigned type, result still unsigned
#ifdef _MSC_VER
#pragma warning(disable: 4146)
#endif


/******************************************************************************
    Node creation
******************************************************************************/

static EXPR* cpp_new_expr(CPP* cpp)
{
    return (EXPR*)cpp_heap_calloc(&cpp->output_heap, 1, sizeof(EXPR));
}


static EXPR* cpp_new_unary_expr(CPP* cpp, int type, EXPR* operand)
{
    EXPR* expr = cpp_new_expr(cpp);

    expr->type = type;
    expr->u.unary.operand = operand;

    return expr;
}


static EXPR* cpp_new_binary_expr(CPP* cpp, EXPR* left, int type, EXPR* right)
{
    EXPR* expr = cpp_new_expr(cpp);

    expr->type = type;
    expr->u.binary.left = left;
    expr->u.binary.right = right;

    return expr;
}


static EXPR* cpp_new_conditional_expr(CPP* cpp, EXPR* condition, EXPR* iftrue, EXPR* iffalse)
{
    EXPR* expr = cpp_new_expr(cpp);

    expr->type = E_CONDITIONAL;
    expr->u.conditional.condition = condition;
    expr->u.conditional.iftrue = iftrue;
    expr->u.conditional.iffalse = iffalse;

    return expr;
}


/******************************************************************************
    Parse value of token
******************************************************************************/

static int cpp_get_digit_value(char ch)
{
    switch ( ch ) {
        case '0': return 0;
        case '1': return 1;
        case '2': return 2;
        case '3': return 3;
        case '4': return 4;
        case '5': return 5;
        case '6': return 6;
        case '7': return 7;
        case '8': return 8;
        case '9': return 9;
        case 'a': case 'A': return 10;
        case 'b': case 'B': return 11;
        case 'c': case 'C': return 12;
        case 'd': case 'D': return 13;
        case 'e': case 'E': return 14;
        case 'f': case 'F': return 15;
    }
    return 16;
}


static void cpp_get_integer_value(CPP* cpp, VALUE* dest, PPTOKEN* token)
{
    PPTOKEN*        error;
    const char*     pos;
    int             digit;
    int             base;

    pos = token->token.str;

    // Handle floating point
    if ( strchr(pos,'.') ) {
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "Floating point not supported by preprocessor: '");
        cpp_msg_cats(cpp, pos);
        cpp_msg_cats(cpp, "'");
        error = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
        cpp_output_append(cpp, error);
        // muddle on anyway...
        dest->is_signed = true;
        dest->u.svalue = atoi(pos);
        return;
    }

    // Determine sign
    if ( strchr(pos,'u') || strchr(pos,'U') ) {
        dest->is_signed = false;
        dest->u.svalue = 0;
    } else {
        dest->is_signed = true;
        dest->u.uvalue = 0;
    }

    // Determine base
    if ( pos[0]=='0' ) {
        if ( pos[1]=='x' || pos[1]=='X' ) {
            base = 16;
            pos += 2;
        } else {
            base = 8;
        }
    } else {
        base = 10;
    }

    // Get number
    while ( (digit = cpp_get_digit_value(*pos)) < base ) {
        if ( dest->is_signed ) {
            dest->u.svalue *= (long)base;
            dest->u.svalue += (long)digit;
        } else {
            dest->u.uvalue *= (unsigned long)base;
            dest->u.uvalue += (unsigned long)digit;
        }
        ++pos;
    }
}


static void cpp_get_character_value(CPP* cpp, VALUE* dest, PPTOKEN* token)
{
    PPTOKEN*        error;
    const char*     pos;
    int             digit;
    size_t          len;
    bool            wide;
    unsigned long   value;

    pos = token->token.str;
    if ( *pos == 'L' ) {
        ++pos;
        wide = true;
        dest->is_signed = false;
    } else {
        wide = false;
        dest->is_signed = true;
    }
    assert(*pos == '\'');
    ++pos;
    dest->is_signed = !wide;

    // Handle empty literal
    if ( *pos == '\'' ) {
        goto bad_literal;
    } 

    if ( *pos != '\\' ) {
        // Simple literal
        if ( dest->is_signed ) {
            dest->u.svalue = (signed char)*pos;
        } else {
            dest->u.uvalue = (unsigned char)*pos;
        }
        ++pos;
    } else if ( (len=cpp_universal_char_len(pos)) != 0 ) {
        // Universal character names
        pos += 2;
        len -= 2;
        value = 0;
        while ( len ) {
            value *= 16;
            value += (unsigned long)cpp_get_digit_value(*pos);
            --len;
            ++pos;
        }
        if ( dest->is_signed ) {
            dest->u.svalue = value;
        } else {
            dest->u.uvalue = value;
        }
    } else if ( pos[1]=='x' && ISXDIGIT(pos[2]) ) {
        // Hexadecimal
        pos += 2;
        value = 0;
        while ( (digit = cpp_get_digit_value(*pos)) < 16 ) {
            value *= 16;
            value += (unsigned long)digit;
            if ( !wide && value > UCHAR_MAX ) {
                goto bad_literal;
            }
            ++pos;
        }
        if ( dest->is_signed ) {
            dest->u.svalue = value;
        } else {
            dest->u.uvalue = value;
        }
    } else if ( ISODIGIT(pos[1]) ) {
        // Octal
        ++pos;
        value = 0;
        len = 0;
        while ( (digit=cpp_get_digit_value(*pos)) < 8 && len < 3 ) {
            value *= 8;
            value += (unsigned long)digit;
            if ( !wide && value > UCHAR_MAX ) {
                goto bad_literal;
            }
            ++pos;
            ++len;
        }
        if ( dest->is_signed ) {
            dest->u.svalue = value;
        } else {
            dest->u.uvalue = value;
        }
    } else {
        // Other escape sequence
        ++pos;
        switch ( *pos++ ) {
            case 'a': value = '\a'; break;
            case 'b': value = '\b'; break;
            case 'f': value = '\f'; break;
            case 'n': value = '\n'; break;
            case 'r': value = '\r'; break;
            case 't': value = '\t'; break;
            case 'v': value = '\v'; break;
            case '\\':
            case '\'':
            case '"':
            case '?': value = pos[1]; break;
            default: goto bad_literal;
        }
        if ( dest->is_signed ) {
            dest->u.svalue = (signed char)*pos;
        } else {
            dest->u.uvalue = (unsigned char)*pos;
        }
    }

    if ( *pos == '\'' ) {
        return;
    }

bad_literal:
    cpp_msg_new(cpp);
    cpp_msg_cats(cpp, "Invalid character: ");
    cpp_msg_cats(cpp, token->token.str);
    error = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
    cpp_output_append(cpp, error);
    if ( dest->is_signed ) {
        dest->u.svalue = ' ';
    } else {
        dest->u.uvalue = ' ';
    }
}


/******************************************************************************
    Expression parser
******************************************************************************/

static EXPR* cpp_parse_conditional_expression(CPP* cpp, PPTOKEN** pos);

static EXPR* cpp_parse_primary_expression(CPP* cpp, PPTOKEN** pos)
{
    EXPR*   expr;

    switch ( (*pos)->token.tag ) {
        case T_NUMLIT:
            expr = cpp_new_expr(cpp);
            expr->type = E_VALUE;
            cpp_get_integer_value(cpp, &expr->u.value, *pos);
            (*pos) = (*pos)->next;
            break;
        case T_CHRLIT:
            expr = cpp_new_expr(cpp);
            expr->type = E_VALUE;
            cpp_get_character_value(cpp, &expr->u.value, *pos);
            (*pos) = (*pos)->next;
            break;
        case T_PAREN_OPEN:
            (*pos) = (*pos)->next;
            expr = cpp_parse_conditional_expression(cpp, pos);
            if ( (*pos)->token.tag != T_PAREN_CLOSE ) {
                expr = NULL;
            } else {
                (*pos) = (*pos)->next;
            }
            break;
        default:
            expr = NULL;
            break;
    }

    return expr;
}


static EXPR* cpp_parse_unary_expression(CPP* cpp, PPTOKEN** pos)
{
    EXPR*   expr;
    int     type;

    switch ( (*pos)->token.tag ) {
        case T_ADD:
            type = E_UPLUS;
            break;
        case T_SUB:
            type = E_UMINUS;
            break;
        case T_NOT:
            type = E_NOT;
            break;
        case T_NEG:
            type = E_NEG;
            break;
        default:
            return cpp_parse_primary_expression(cpp, pos);
    }

    (*pos) = (*pos)->next;
    expr = cpp_parse_unary_expression(cpp, pos);
    return expr
        ? cpp_new_unary_expr(cpp, type, expr)
        : NULL;
}


static EXPR* cpp_parse_multiplicative_expression(CPP* cpp, PPTOKEN** pos)
{
    EXPR*   left;
    EXPR*   right;
    int     type;

    left = cpp_parse_unary_expression(cpp, pos);
    if ( left == NULL ) {
        return left;
    }
    switch ( (*pos)->token.tag ) {
        case T_MUL:
            type = E_MUL;
            break;
        case T_DIV:
            type = E_DIV;
            break;
        case T_MOD:
            type = E_MOD;
            break;
        default:
            return left;
    }
    (*pos) = (*pos)->next;
    right = cpp_parse_multiplicative_expression(cpp, pos);
    return right
        ? cpp_new_binary_expr(cpp, left, type, right)
        : NULL;
}


static EXPR* cpp_parse_additive_expression(CPP* cpp, PPTOKEN** pos)
{
    EXPR*   left;
    EXPR*   right;
    int     type;

    left = cpp_parse_multiplicative_expression(cpp, pos);
    if ( left == NULL ) {
        return left;
    }
    switch ( (*pos)->token.tag ) {
        case T_ADD:
            type = E_ADD;
            break;
        case T_SUB:
            type = E_SUB;
            break;
        default:
            return left;
    }
    (*pos) = (*pos)->next;
    right = cpp_parse_additive_expression(cpp, pos);
    return right
        ? cpp_new_binary_expr(cpp, left, type, right)
        : NULL;
}


static EXPR* cpp_parse_shift_expression(CPP* cpp, PPTOKEN** pos)
{
    EXPR*   left;
    EXPR*   right;
    int     type;

    left = cpp_parse_additive_expression(cpp, pos);
    if ( left == NULL ) {
        return left;
    }
    switch ( (*pos)->token.tag ) {
        case T_SHL:
            type = E_SHL;
            break;
        case T_SHR:
            type = E_SHR;
            break;
        default:
            return left;
    }
    (*pos) = (*pos)->next;
    right = cpp_parse_shift_expression(cpp, pos);
    return right
        ? cpp_new_binary_expr(cpp, left, type, right)
        : NULL;
}


static EXPR* cpp_parse_relational_expression(CPP* cpp, PPTOKEN** pos)
{
    EXPR*   left;
    EXPR*   right;
    int     type;

    left = cpp_parse_shift_expression(cpp, pos);
    if ( left == NULL ) {
        return left;
    }
    switch ( (*pos)->token.tag ) {
        case T_GT:
            type = E_GT;
            break;
        case T_GE:
            type = E_GE;
            break;
        case T_LT:
            type = E_LT;
            break;
        case T_LE:
            type = E_LE;
            break;
        default:
            return left;
    }
    (*pos) = (*pos)->next;
    right = cpp_parse_relational_expression(cpp, pos);
    return right
        ? cpp_new_binary_expr(cpp, left, type, right)
        : NULL;
}


static EXPR* cpp_parse_equality_expression(CPP* cpp, PPTOKEN** pos)
{
    EXPR*   left;
    EXPR*   right;
    int     type;

    left = cpp_parse_relational_expression(cpp, pos);
    if ( left == NULL ) {
        return left;
    }
    switch ( (*pos)->token.tag ) {
        case T_EQ:
            type = E_EQ;
            break;
        case T_NE:
            type = E_NE;
            break;
        default:
            return left;
    }
    (*pos) = (*pos)->next;
    right = cpp_parse_equality_expression(cpp, pos);
    return right
        ? cpp_new_binary_expr(cpp, left, type, right)
        : NULL;
}


static EXPR* cpp_parse_bit_and_expression(CPP* cpp, PPTOKEN** pos)
{
    EXPR*   left;
    EXPR*   right;

    left = cpp_parse_equality_expression(cpp, pos);
    if ( left == NULL || (*pos)->token.tag != T_AND ) {
        return left;
    }
    (*pos) = (*pos)->next;
    right = cpp_parse_bit_and_expression(cpp, pos);
    return right
        ? cpp_new_binary_expr(cpp, left, E_BIT_AND, right)
        : NULL;
}


static EXPR* cpp_parse_bit_xor_expression(CPP* cpp, PPTOKEN** pos)
{
    EXPR*   left;
    EXPR*   right;

    left = cpp_parse_bit_and_expression(cpp, pos);
    if ( left == NULL || (*pos)->token.tag != T_XOR ) {
        return left;
    }
    (*pos) = (*pos)->next;
    right = cpp_parse_bit_xor_expression(cpp, pos);
    return right
        ? cpp_new_binary_expr(cpp, left, E_BIT_XOR, right)
        : NULL;
}


static EXPR* cpp_parse_bit_or_expression(CPP* cpp, PPTOKEN** pos)
{
    EXPR*   left;
    EXPR*   right;

    left = cpp_parse_bit_xor_expression(cpp, pos);
    if ( left == NULL || (*pos)->token.tag != T_OR ) {
        return left;
    }
    (*pos) = (*pos)->next;
    right = cpp_parse_bit_or_expression(cpp, pos);
    return right
        ? cpp_new_binary_expr(cpp, left, E_BIT_OR, right)
        : NULL;
}


static EXPR* cpp_parse_logical_and_expression(CPP* cpp, PPTOKEN** pos)
{
    EXPR*   left;
    EXPR*   right;

    left = cpp_parse_bit_or_expression(cpp, pos);
    if ( left == NULL || (*pos)->token.tag != T_AND_AND ) {
        return left;
    }
    (*pos) = (*pos)->next;
    right = cpp_parse_logical_and_expression(cpp, pos);
    return right
        ? cpp_new_binary_expr(cpp, left, E_LOGIC_AND, right)
        : NULL;
}


static EXPR* cpp_parse_logical_or_expression(CPP* cpp, PPTOKEN** pos)
{
    EXPR*   left;
    EXPR*   right;

    left = cpp_parse_logical_and_expression(cpp, pos);
    if ( left == NULL || (*pos)->token.tag != T_OR_OR ) {
        return left;
    }
    (*pos) = (*pos)->next;
    right = cpp_parse_logical_or_expression(cpp, pos);
    return right
        ? cpp_new_binary_expr(cpp, left, E_LOGIC_OR, right)
        : NULL;
}


static EXPR* cpp_parse_conditional_expression(CPP* cpp, PPTOKEN** pos)
{
    EXPR* expr;
    EXPR* iftrue;
    EXPR* iffalse;
    
    expr = cpp_parse_logical_or_expression(cpp, pos);
    if ( expr == NULL || (*pos)->token.tag != T_QMARK ) {
        return expr;
    }
    (*pos) = (*pos)->next;
    iftrue = cpp_parse_conditional_expression(cpp, pos);
    if ( iftrue==NULL || (*pos)->token.tag != T_COLON ) {
        return NULL;
    }
    (*pos) = (*pos)->next;
    iffalse = cpp_parse_conditional_expression(cpp, pos);
    if ( iffalse == NULL ) {
        return NULL;
    }
    return cpp_new_conditional_expr(cpp, expr, iftrue, iffalse);
}


/******************************************************************************
    Expression evaluation
******************************************************************************/

static bool cpp_is_zero(VALUE* value)
{
    return value->is_signed
        ? ( value->u.svalue == 0 )
        : ( value->u.uvalue == 0 );
}


static void cpp_set_zero(VALUE* value)
{
    if ( value->is_signed ) {
        value->u.svalue = 0;
    } else {
        value->u.uvalue = 0;
    }
}


static void cpp_adjust_sign(VALUE* left, VALUE* right)
{
    if ( !left->is_signed ) {
        if ( right->is_signed ) {
            right->u.uvalue = (unsigned long)right->u.svalue;
            right->is_signed = false;
        }
    } else if ( !right->is_signed ) {
        if ( left->is_signed ) {
            left->u.uvalue = (unsigned long)left->u.svalue;
            left->is_signed = false;
        }
    }
}


static void cpp_eval_expression(CPP* cpp, VALUE* result, EXPR* expr)
{
    VALUE   left;
    VALUE   right;

    switch ( expr->type ) {
        case E_VALUE:
            *result = expr->u.value;
            break;
        case E_CONDITIONAL:
            cpp_eval_expression(cpp, &left, expr->u.conditional.condition);
            if ( !cpp_is_zero(&left) ) {
                cpp_eval_expression(cpp, result, expr->u.conditional.iftrue);
            } else {
                cpp_eval_expression(cpp, result, expr->u.conditional.iffalse);
            }
            break;
        case E_LOGIC_OR:
            result->is_signed = true;
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            if ( !cpp_is_zero(&left) ) {
                result->u.svalue = 1;
            } else {
                cpp_eval_expression(cpp, &right, expr->u.binary.right);
                result->u.svalue = !cpp_is_zero(&right);
            }
            break;
        case E_LOGIC_AND:
            result->is_signed = true;
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            if ( cpp_is_zero(&left) ) {
                result->u.svalue = 0;
            } else {
                cpp_eval_expression(cpp, &right, expr->u.binary.right);
                result->u.svalue = !cpp_is_zero(&right);
            }
            break;
        case E_BIT_OR:
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            cpp_eval_expression(cpp, &right, expr->u.binary.right);
            cpp_adjust_sign(&left, &right);
            result->is_signed = left.is_signed;
            if ( result->is_signed ) {
                result->u.svalue = left.u.svalue | right.u.svalue;
            } else {
                result->u.uvalue = left.u.uvalue | right.u.uvalue;
            }
            break;
        case E_BIT_XOR:
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            cpp_eval_expression(cpp, &right, expr->u.binary.right);
            cpp_adjust_sign(&left, &right);
            result->is_signed = left.is_signed;
            if ( result->is_signed ) {
                result->u.svalue = left.u.svalue ^ right.u.svalue;
            } else {
                result->u.uvalue = left.u.uvalue ^ right.u.uvalue;
            }
            break;
        case E_BIT_AND:
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            cpp_eval_expression(cpp, &right, expr->u.binary.right);
            cpp_adjust_sign(&left, &right);
            result->is_signed = left.is_signed;
            if ( result->is_signed ) {
                result->u.svalue = left.u.svalue & right.u.svalue;
            } else {
                result->u.uvalue = left.u.uvalue & right.u.uvalue;
            }
            break;
        case E_SHL:
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            cpp_eval_expression(cpp, &right, expr->u.binary.right);
            result->is_signed = left.is_signed;
            if ( left.is_signed ) {
                if ( right.is_signed ) {
                    result->u.svalue = left.u.svalue << right.u.svalue;
                } else {
                    result->u.svalue = left.u.svalue << right.u.uvalue;
                }
            } else {
                if ( right.is_signed ) {
                    result->u.uvalue = left.u.uvalue << right.u.svalue;
                } else {
                    result->u.uvalue = left.u.uvalue << right.u.uvalue;
                }
            }
            break;
        case E_SHR:
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            cpp_eval_expression(cpp, &right, expr->u.binary.right);
            result->is_signed = left.is_signed;
            if ( left.is_signed ) {
                if ( right.is_signed ) {
                    result->u.svalue = left.u.svalue >> right.u.svalue;
                } else {
                    result->u.svalue = left.u.svalue >> right.u.uvalue;
                }
            } else {
                if ( right.is_signed ) {
                    result->u.uvalue = left.u.uvalue >> right.u.svalue;
                } else {
                    result->u.uvalue = left.u.uvalue >> right.u.uvalue;
                }
            }
            break;
        case E_EQ:
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            cpp_eval_expression(cpp, &right, expr->u.binary.right);
            cpp_adjust_sign(&left, &right);
            result->is_signed = true;
            result->u.svalue = left.is_signed 
                ? ( left.u.svalue == right.u.svalue )
                : ( left.u.uvalue == right.u.uvalue );
            break;
        case E_NE:
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            cpp_eval_expression(cpp, &right, expr->u.binary.right);
            cpp_adjust_sign(&left, &right);
            result->is_signed = true;
            result->u.svalue = left.is_signed 
                ? ( left.u.svalue != right.u.svalue )
                : ( left.u.uvalue != right.u.uvalue );
            break;
        case E_GT:
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            cpp_eval_expression(cpp, &right, expr->u.binary.right);
            cpp_adjust_sign(&left, &right);
            result->is_signed = true;
            result->u.svalue = left.is_signed 
                ? ( left.u.svalue > right.u.svalue )
                : ( left.u.uvalue > right.u.uvalue );
            break;
        case E_GE:
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            cpp_eval_expression(cpp, &right, expr->u.binary.right);
            cpp_adjust_sign(&left, &right);
            result->is_signed = true;
            result->u.svalue = left.is_signed 
                ? ( left.u.svalue >= right.u.svalue )
                : ( left.u.uvalue >= right.u.uvalue );
            break;
        case E_LT:
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            cpp_eval_expression(cpp, &right, expr->u.binary.right);
            cpp_adjust_sign(&left, &right);
            result->is_signed = true;
            result->u.svalue = left.is_signed 
                ? ( left.u.svalue < right.u.svalue )
                : ( left.u.uvalue < right.u.uvalue );
            break;
        case E_LE:
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            cpp_eval_expression(cpp, &right, expr->u.binary.right);
            cpp_adjust_sign(&left, &right);
            result->is_signed = true;
            result->u.svalue = left.is_signed 
                ? ( left.u.svalue <= right.u.svalue )
                : ( left.u.uvalue <= right.u.uvalue );
            break;
        case E_ADD:
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            cpp_eval_expression(cpp, &right, expr->u.binary.right);
            cpp_adjust_sign(&left, &right);
            result->is_signed = left.is_signed;
            if ( left.is_signed ) {
                result->u.svalue = left.u.svalue + right.u.svalue;
            } else {
                result->u.uvalue = left.u.uvalue + right.u.uvalue;
            }
            break;
        case E_SUB:
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            cpp_eval_expression(cpp, &right, expr->u.binary.right);
            cpp_adjust_sign(&left, &right);
            result->is_signed = left.is_signed;
            if ( left.is_signed ) {
                result->u.svalue = left.u.svalue - right.u.svalue;
            } else {
                result->u.uvalue = left.u.uvalue - right.u.uvalue;
            }
            break;
        case E_MUL:
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            cpp_eval_expression(cpp, &right, expr->u.binary.right);
            cpp_adjust_sign(&left, &right);
            result->is_signed = left.is_signed;
            if ( left.is_signed ) {
                result->u.svalue = left.u.svalue * right.u.svalue;
            } else {
                result->u.uvalue = left.u.uvalue * right.u.uvalue;
            }
            break;
        case E_DIV:
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            cpp_eval_expression(cpp, &right, expr->u.binary.right);
            cpp_adjust_sign(&left, &right);
            result->is_signed = left.is_signed;
            if ( cpp_is_zero(&right) ) {
                cpp_output_append(cpp, cpp_token_news(cpp, T_ERROR, "Division by zero"));
                cpp_set_zero(result);
                break;
            }
            if ( left.is_signed ) {
                result->u.svalue = left.u.svalue / right.u.svalue;
            } else {
                result->u.uvalue = left.u.uvalue / right.u.uvalue;
            }
            break;
        case E_MOD:
            cpp_eval_expression(cpp, &left, expr->u.binary.left);
            cpp_eval_expression(cpp, &right, expr->u.binary.right);
            cpp_adjust_sign(&left, &right);
            result->is_signed = left.is_signed;
            if ( cpp_is_zero(&right) ) {
                cpp_output_append(cpp, cpp_token_news(cpp, T_ERROR, "Modulus by zero"));
                cpp_set_zero(result);
                break;
            }
            if ( left.is_signed ) {
                result->u.svalue = left.u.svalue % right.u.svalue;
            } else {
                result->u.uvalue = left.u.uvalue % right.u.uvalue;
            }
            break;
        case E_UPLUS:
            cpp_eval_expression(cpp, result, expr->u.unary.operand);
            break;
        case E_UMINUS:
            cpp_eval_expression(cpp, &right, expr->u.unary.operand);
            result->is_signed = right.is_signed;
            if ( result->is_signed ) {
                result->u.svalue = - right.u.svalue;
            } else {
                result->u.uvalue = - right.u.uvalue;
            }
            break;
        case E_NOT:
            cpp_eval_expression(cpp, &right, expr->u.unary.operand);
            result->is_signed = true;
            result->u.svalue = cpp_is_zero(&right);
            break;
        case E_NEG:
            cpp_eval_expression(cpp, &right, expr->u.unary.operand);
            if ( right.is_signed ) {
                result->is_signed = true;
                result->u.svalue = ~right.u.svalue;
            } else {
                result->is_signed = false;
                result->u.svalue = ~right.u.svalue;
            }
            break;
        default:
            assert(0);
            break;
    }
}


/******************************************************************************
    Evaluate 'defined' operators
******************************************************************************/

static PPTOKEN* cpp_bad_defined(CPP* cpp, const char* ident)
{
    cpp_msg_new(cpp);
    cpp_msg_cats(cpp, "Syntax error in #");
    cpp_msg_cats(cpp, ident);
    cpp_msg_cats(cpp, " expression near 'defined'");
    cpp_output_append(cpp, cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp)));
    return cpp_token_news(cpp, T_NUMLIT, "0");
}


static PPTOKEN* cpp_eval_defined(CPP* cpp, const char* ident, PPTOKEN* list)
{
    PPLIST      result;
    PPTOKEN*    token;
    const char* truth;
    int         paren;

    cpp_list_reset(&result);
    while ( list ) {
        if ( list->token.tag != T_IDENT || strcmp(list->token.str,"defined") ) {
            token = list;
            list = list->next;
            cpp_list_append(&result, token);
        } else {
            paren = 0;
            list = list->next;
            while ( list && (list->token.tag == T_SPACE || list->token.tag == T_PAREN_OPEN) ) {
                if ( list->token.tag == T_PAREN_OPEN ) {
                    ++paren;
                }
                list = list->next;
            }
            if ( list == 0 || list->token.tag != T_IDENT ) {
                return cpp_bad_defined(cpp, ident);
            }
            truth = cpp_find_macro(cpp, list->token.str) ? "1" : "0";
            while ( paren-- ) {
                list = cpp_skip_t_space(list->next);
                if ( list == NULL || list->token.tag != T_PAREN_CLOSE ) {
                    return cpp_bad_defined(cpp, ident);
                }
            }
            cpp_list_append(&result, cpp_token_news(cpp, T_NUMLIT, truth));
            list = list->next;
        }
    }

    return result.first;
}


/******************************************************************************
    #if/#elif directive evaluation
******************************************************************************/

bool cpp_eval_if_expression(CPP* cpp, const char* ident, const char* pos)
{
    PPLIST      nospace;
    PPTOKEN*    list;
    PPTOKEN*    token;
    EXPR*       expr;
    VALUE       value;

    // Convert line to list of tokens
    list = cpp_tokenize(cpp, pos);

    // Identify 'defined' and change tag so it does not get replaced
    list = cpp_eval_defined(cpp, ident, list);

    // Expand the list
    list = cpp_expand_list(cpp, list, NULL);

    // Remove spaces to make it easier to parse
    cpp_list_reset(&nospace);
    while ( list ) {
        token = list;
        list = list->next;
        if ( token->token.tag != T_SPACE ) {
            cpp_list_append(&nospace, token);
        }
    }

    // Handle empty expression
    if ( nospace.first == NULL ) {
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "Expression expected after '#");
        cpp_msg_cats(cpp, ident);
        cpp_msg_cats(cpp, "'");
        token = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
        cpp_output_append(cpp, token);
        return false;
    }

    // Add a T_END token to avoid NULL pointer checks
    cpp_list_append(&nospace, cpp_token_new(cpp, T_END));

    // Convert remaining identifiers to 0
    for ( token = nospace.first ; token ; token = token->next ) {
        if ( token->token.tag == T_IDENT || token->token.tag == T_FINAL_IDENT ) {
            token->token.tag = T_NUMLIT;
            token->token.str = "0";
        }
    }

    // Parse expression
    token = nospace.first;
    expr = cpp_parse_conditional_expression(cpp, &token);
    if ( expr==NULL || token->token.tag != T_END ) {
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "Syntax error in #");
        cpp_msg_cats(cpp, ident);
        cpp_msg_cats(cpp, " expression");
        if ( token->token.tag == T_END ) {
            cpp_msg_cats(cpp, " at end of line");
        } else {
            cpp_msg_cats(cpp, " near '");
            cpp_msg_cats(cpp, token->token.str);
            cpp_msg_cats(cpp, "'");
        }
        token = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
        cpp_output_append(cpp, token);
        return false;
    }

    // Evaluate expression
    cpp_eval_expression(cpp, &value, expr);

    return !cpp_is_zero(&value);
}
