/******************************************************************************

    $Id: heap.c,v 1.1 2003/08/28 05:46:47 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/heap.c,v $

******************************************************************************/

#include "priv.h"

// Forwards.
typedef union tagALIGNMENT ALIGNMENT;


// To make alignment right
union tagALIGNMENT {
    long            l;
    double          d;
    void*           p;
};


// Block allocated from system
struct tagSYSMEM {
    SYSMEM*         next;               // Next in list
    size_t          size;               // Allocated size
    size_t          used;               // Number of bytes used
    union {
        ALIGNMENT   align;              // To make alignment right
        char        data[1];            // Allocated using malloc()
    } u;
};


void cpp_heap_init(CPP* cpp, HEAP* heap)
{
    memset(heap, 0, sizeof(HEAP));
    heap->cpp = cpp;
}


void cpp_heap_cleanup(CPP* cpp)
{
    SYSMEM* sysmem;

    while ( (sysmem=cpp->sysmem_freelist) != NULL ) {
        cpp->sysmem_freelist = sysmem->next;
        cpp_free(cpp, sysmem);
    }
}


void cpp_heap_freetop(HEAP* heap)
{
    SYSMEM* sysmem = heap->top;

    heap->size -= sysmem->size;
    heap->top = sysmem->next;
    if ( sysmem->size != SYSMEM_BLOCKSIZE ) {
        cpp_free(heap->cpp, sysmem);
    } else {
        sysmem->next = heap->cpp->sysmem_freelist;
        heap->cpp->sysmem_freelist = sysmem;
    }
}


void cpp_heap_clear(HEAP* heap)
{
    while ( heap->top ) {
        cpp_heap_freetop(heap);
    }
}


void* cpp_heap_malloc(HEAP* heap, size_t size)
{
    SYSMEM* sysmem = heap->top;
    char*   ptr;
    size_t  remainder;

    remainder = size % sizeof(ALIGNMENT);
    if (remainder) {
        size += sizeof(ALIGNMENT) - remainder;
    }

    if ( sysmem == NULL || size > sysmem->size - sysmem->used ) {
        if ( size <= SYSMEM_BLOCKSIZE && heap->cpp->sysmem_freelist ) {
            sysmem = heap->cpp->sysmem_freelist;
            heap->cpp->sysmem_freelist = sysmem->next;
        } else {
            size_t n = SYSMEM_BLOCKSIZE;
            if ( n < size ) {
                n = size;
            }
            sysmem = cpp_safe_malloc(heap->cpp, sizeof(SYSMEM)-sizeof(ALIGNMENT)+n);
            sysmem->size = n;
        }
        sysmem->next = heap->top;
        sysmem->used = 0;
        heap->size += sysmem->size;
        if ( heap->size > heap->top_size ) {
            heap->top_size = heap->size;
        }
        heap->top = sysmem;
    }

    ptr = sysmem->u.data + sysmem->used;
    sysmem->used += size;

    return ptr;
}


void* cpp_heap_calloc(HEAP* heap, size_t num, size_t size)
{
    size_t bytes = num * size;
    void* ptr;

    if ( bytes ) {
        ptr = cpp_heap_malloc(heap, bytes);
        memset(ptr, 0, bytes);
    } else {
        ptr = NULL;
    }

    return ptr;
}


void* cpp_heap_mark(HEAP* heap)
{
    return cpp_heap_malloc(heap, 0);
}


void cpp_heap_release(HEAP* heap, void* mark)
{
    SYSMEM* top;
    char*   cmark = (char*)mark;

    for (;;) {
        top = heap->top;
        if ( cmark >= top->u.data && cmark <= top->u.data + top->size ) {
            top->used = top->size - (cmark - top->u.data);
            if ( top->used == 0 ) {
                cpp_heap_freetop(heap);
            }
            return;
        }
        cpp_heap_freetop(heap);
    }
}


void* cpp_heap_strndup(HEAP* heap, const char* str, size_t len)
{
    char* result = (char*)cpp_heap_malloc(heap, len+1);

    memcpy(result, str, len);
    result[len] = '\0';

    return result;
}


void* cpp_heap_strdup(HEAP* heap, const char* str)
{
    return cpp_heap_strndup(heap, str, strlen(str));
}
