/******************************************************************************

    $Id: input.c,v 1.3 2003/08/28 09:53:28 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/input.c,v $

******************************************************************************/

#include "priv.h"

/******************************************************************************
    Local functions
******************************************************************************/

// Read logical line
static void cpp_read_logical_line(CPP* cpp)
{
    static const char tri_find[] = "=(/)'<!>-";
    static const char tri_repl[] = "#[\\]^{|}~";

    PPTOKEN*    token;
    FILE*       fp = cpp->input->fp;
    char        ch = '\n';
    int         i;
    size_t      len;
    size_t      before;
    size_t      njoin = 0;
    char*       tri;
    const char* pos;

    for (;;) {
        len = cpp->input_buf.len;
        // Read physical line.
        for (;;) {
            if ( (i=fgetc(fp)) == EOF ) {
                if ( ferror(fp) ) {
                    cpp_msg_new(cpp);
                    cpp_msg_cats(cpp, "Error reading \"");
                    cpp_msg_cats(cpp, cpp->input->path);
                    cpp_msg_cats(cpp, "\": ");
                    cpp_msg_cats(cpp, strerror(errno));
                    cpp_fatal(cpp, cpp_msg_gets(cpp));
                }
                if ( ch != '\n' ) {
                    token = cpp_token_news(cpp, T_ERROR, "Missing newline at end-of-file");
                    cpp_output_append(cpp, token);
                }
                break;
            }
            ch = (char)i;
            if ( ch == '\0' ) {
                cpp_msg_new(cpp);
                cpp_msg_cats(cpp, "NUL character found in \"");
                cpp_msg_cats(cpp, cpp->input->path);
                cpp_msg_cats(cpp, "\"");
                cpp_fatal(cpp, cpp_msg_gets(cpp));
            }
            cpp_buffer_catc(&cpp->input_buf, ch);
            if ( ch == '\n' ) {
                ++(cpp->input->nlcount);
                break;
            }
        }

        // Process trigraphs
        tri = strstr(cpp_buffer_gets(&cpp->input_buf)+len, "?" "?");
        while ( tri ) {
            ch = tri[2];
            if ( ch && (pos=strchr(tri_find,ch)) != NULL  ) {
                before = tri - cpp->input_buf.data;
                tri[0] = tri_repl[pos - tri_find];
                memmove(tri+1, tri+3, cpp->input_buf.len - before - 1);
                cpp->input_buf.len -= 2;
            }
            tri = strstr(tri+1, "?" "?");
        }

        // Handle line continuation
        pos = cpp->input_buf.data + cpp->input_buf.len;
        if ( cpp->input_buf.len < len + 2 || pos[-1] != '\n' || pos[-2] != '\\' ) {
            break;
        }
        cpp->input_buf.len -= 2;
        ++njoin;
    }

    while ( njoin-- ) {
        cpp_buffer_catc(&cpp->input_buf, '\n');
    }
}


static void cpp_open_quote(CPP* cpp, char quote)
{
    cpp_msg_new(cpp);
    cpp_msg_cats(cpp, "Open quote - missing ");
    cpp_msg_catc(cpp, quote);
    cpp_output_append(cpp, cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp)));
}


// Read logical line and make comments regular white space
static void cpp_read_uncommented_line(CPP* cpp)
{
    PPTOKEN*    token;
    char*       cp;
    char        quote;
    size_t      len;
    size_t      pos;

    pos = cpp->input_buf.len;
    cpp_read_logical_line(cpp);
    cp = cpp_buffer_gets(&cpp->input_buf) + pos;

    while ( *cp ) {
        switch ( *cp ) {
            case '"':
            case '\'':
                quote = *cp++;
                while ( *cp != quote ) {
                    if ( cp[0]=='\n' || cp[0]=='\0' ) {
                        // Report problem
                        cpp_open_quote(cpp, quote);
                        // Fix the problem
                        pos = cp - cpp->input_buf.data;
                        cpp_buffer_insertc(&cpp->input_buf, pos, quote);
                        cp = cpp_buffer_gets(&cpp->input_buf) + pos + 1;
                        break;
                    }
                    if ( *cp == '\\' ) {
                        if ( cp[1] == '\n' || cp[1] == '\0' ) {
                            cpp_open_quote(cpp, quote);
                            *cp = quote;
                            break;
                        }
                        ++cp;
                    }
                    ++cp;
                }
                ++cp;
                break;
            case '/':
                switch ( cp[1] ) {
                    case '/':
                        while ( *cp && *cp != '\n' ) {
                            if ( !ISSPACE(*cp) ) {
                                *cp = ' ';
                            }
                            ++cp;
                        }
                        break;
                    case '*':
                        *cp++ = ' ';
                        *cp++ = ' ';
                        while ( cp[0] != '*' || cp[1] != '/' ) {
                            if ( cp[0] == '\0' ) {
                                len = cpp->input_buf.len;
                                cpp_read_logical_line(cpp);
                                if ( len == cpp->input_buf.len ) {
                                    cpp_msg_new(cpp);
                                    cpp_msg_cats(cpp, "'/*' has no '*/'");
                                    token = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
                                    cpp_output_append(cpp, token);
                                    return;
                                }
                                cp = cpp_buffer_gets(&cpp->input_buf) + len;
                            } else {
                                if ( !ISSPACE(*cp) ) {
                                    *cp = ' ';
                                }
                                ++cp;
                            }
                        }
                        *cp++ = ' ';
                        *cp++ = ' ';
                        break;
                    default:
                        ++cp;
                        break;
                }
                break;
            default:
                ++cp;
                break;
        }
    }
}


/******************************************************************************
    Entry points
******************************************************************************/

void cpp_input_push(CPP* cpp, const char* path)
{
    INPUT* input;

    if ( cpp->input_freelist ) {
        input = cpp->input_freelist;
        cpp->input_freelist = input->parent;
    } else {
        input = cpp_heap_malloc(&cpp->unit_heap, sizeof(INPUT));
    }

    memset(input, 0, sizeof(INPUT));
    if ( cpp->input ) {
        input->cond_top = cpp->input->cond_top;
    }
    input->parent = cpp->input;
    input->path = cpp_heap_strdup(&cpp->unit_heap, path);
    input->file = input->path;
    cpp->file = input->file;
    cpp->line = 1;
    cpp->input = input;
}


void cpp_input_pop(CPP* cpp)
{
    INPUT* input = cpp->input;

    if ( input->fp && input->owner ) {
        fclose(input->fp);
    }

    cpp->input = input->parent;
    input->parent = cpp->input_freelist;
    cpp->input_freelist = input;

    if ( cpp->input ) {
        cpp->file = cpp->input->file;
        cpp->line = cpp->input->nlcount + 1;
    }
}


void cpp_input_popall(CPP* cpp)
{
    while ( cpp->input ) {
        cpp_input_pop(cpp);
    }
}


PPTOKEN* cpp_input_entok(CPP* cpp)
{
    const char* pos;
    PPTOKEN*    token;
    int         tag;
    size_t      len;

    // Handle pushed back tokens
    if ( cpp->push_back ) {
        token = cpp->push_back;
        cpp->push_back = NULL;
        return token;
    }

    // Ensure to have text to tokenize
    if ( *cpp->input_ptr == '\0' ) {
        cpp->line = cpp->input->nlcount + 1;
        cpp_buffer_clear(&cpp->input_buf);
        cpp_read_uncommented_line(cpp);
        cpp->input_ptr = cpp_buffer_gets(&cpp->input_buf);
        if ( *cpp->input_ptr == '\0' ) {
            return cpp_token_new(cpp, T_END);
        }
        pos = cpp_skipspace(cpp->input_ptr);
        cpp_identify_token(pos, &tag, &len);
        if ( tag == T_HASH ) {
            token = cpp_token_newn(cpp, T_BOL_HASH, pos, len);
            cpp->input_ptr = pos + len;
            return token;
        }
    }

    // Extract and return next token
    cpp_identify_token(cpp->input_ptr, &tag, &len);
    token = cpp_token_newn(cpp, tag, cpp->input_ptr, len);
    cpp->input_ptr += len;

    return token;
}
