/******************************************************************************

    $Id: macro.c,v 1.5 2003/08/28 21:59:29 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/macro.c,v $

******************************************************************************/

#include "priv.h"


// Special macro numbers
enum {
    NOT_SPECIAL,
    BUILTIN_LINE,
    BUILTIN_FILE
};


// Compare two macros for equality
bool cpp_is_macro_equal(MACRO* macro1, MACRO* macro2)
{
    size_t      argno;
    MACROTOKEN* body1;
    MACROTOKEN* body2;

    if ( macro1->fnlike != macro2->fnlike || macro1->argc != macro2->argc ) {
        return false;
    }

    for ( argno = 0 ; argno < macro1->argc ; argno++ ) {
        if ( strcmp(macro1->argv[argno], macro2->argv[argno]) ) {
            return false;
        }
    }

    body1 = macro1->body;
    body2 = macro2->body;
    while ( body1 && body2 ) {
        if ( strcmp(body1->str, body2->str) ) {
            return false;
        }
        body1 = body1->next;
        body2 = body2->next;
    }

    return body1==NULL && body2==NULL;
}


// Test if macro name is reserved
bool cpp_is_macro_reserved(const char* name)
{
    if ( name[0] != '_' ) {
        return strcmp(name,"defined")==0;
    } else if ( name[1] != '_' ) {
        return false;
    } else {
        return strcmp(name, "__LINE__") == 0
            || strcmp(name, "__FILE__") == 0
            || strcmp(name, "__DATE__") == 0
            || strcmp(name, "__TIME__") == 0
            || strcmp(name, "__STDC__") == 0
            || strcmp(name, "__STDC_VERSION__") == 0
            || strcmp(name, "__STDC_IEC_559__") == 0
            || strcmp(name, "__STDC_ISO_10646__") == 0
            || strcmp(name, "__STDC_IEC_559_COMPLEX__") == 0;
    }
}


// Look up macro
MACRO* cpp_find_macro(CPP* cpp, const char* name)
{
    MACRO*  macro;
    size_t  h;
    size_t  m;

    h = cpp_hash(name);
    m = h % cpp->macro_hsize;
    for ( macro = cpp->macro_htab[m] ; macro ; macro = macro->next ) {
        if ( strcmp(name, macro->name) == 0 ) {
            return macro;
        }
    }

    return NULL;
}


// Undefine macro if it exists
void cpp_undefine_macro(CPP* cpp, const char* name)
{
    MACRO** macrop;
    MACRO*  macro;
    size_t  h;
    size_t  m;

    h = cpp_hash(name);
    m = h % cpp->macro_hsize;
    for ( macrop = &cpp->macro_htab[m] ; (macro=*macrop) != NULL ; macrop = &macro->next ) {
        if ( strcmp(macro->name, name) == 0 ) {
            *macrop = macro->next;
            cpp->macro_count--;
            break;
        }
    }
}


// Define new macro
void cpp_define_macro(CPP* cpp, MACRO* macro)
{
    MACRO*  old;
    size_t  h;
    size_t  m;

    // Try to find existing macro
    h = cpp_hash(macro->name);
    m = h % cpp->macro_hsize;
    for ( old = cpp->macro_htab[m] ; old ; old = old->next ) {
        if ( strcmp(macro->name, old->name) == 0 ) {
            if ( !cpp_is_macro_equal(macro,old) ) {
                PPTOKEN* token;
                cpp_msg_new(cpp);
                cpp_msg_cats(cpp, "Redefinition of macro '");
                cpp_msg_cats(cpp, macro->name);
                cpp_msg_cats(cpp, "' first defined at ");
                cpp_msg_catfl(cpp, old->file, old->line);
                token = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
                cpp_output_append(cpp, token);
            }
            return;
        }
    }

    // Insert into hash table
    macro->next = cpp->macro_htab[m];
    cpp->macro_htab[m] = macro;
    cpp->macro_count++;

    // Rehashing
    if ( cpp->macro_count > cpp->macro_hsize / 2 ) {
        size_t macro_hsize = 2 * (cpp->macro_hsize + 1) - 1;
        MACRO** macro_htab = (MACRO**)cpp_safe_calloc(cpp, macro_hsize, sizeof(MACRO*));
        MACRO* tmp;
        size_t i;
        for ( i = 0 ; i < cpp->macro_hsize ; i++ ) {
            while ( (tmp=cpp->macro_htab[i]) != NULL ) {
                cpp->macro_htab[i] = tmp->next;
                h = cpp_hash(tmp->name) % macro_hsize;
                tmp->next = macro_htab[h];
                macro_htab[h] = tmp;
            }
        }
        cpp_free(cpp, cpp->macro_htab);
        cpp->macro_htab = macro_htab;
        cpp->macro_hsize = macro_hsize;
    }
}


// Parse '#define' directive
void cpp_parse_define(CPP* cpp, const char* pos)
{
    MACRO*      macro;
    MACROTOKEN* mtoken;
    MACROTOKEN* rev_body;
    PPTOKEN*    pptoken;
    int         tag;
    int         prev_tag;
    size_t      len;
    char**      argv;
    size_t      argv_size;
    size_t      bytes;
    bool        space;
    const char* msg;

    // Get name of macro to define
    pos = cpp_skipspace(pos);
    cpp_identify_token(pos, &tag, &len);
    if ( tag != T_IDENT ) {
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "Expected identifier after '#define'");
        if ( tag != T_END ) {
            cpp_msg_cats(cpp, ", found ");
            cpp_msg_cats(cpp, cpp_strtag(tag));
        }
        pptoken = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
        cpp_output_append(cpp, pptoken);
        return;
    }

    // Create new macro
    macro = cpp_heap_calloc(&cpp->unit_heap, 1, sizeof(MACRO));
    macro->name = cpp_heap_strndup(&cpp->unit_heap, pos, len);
    macro->file = cpp->file;
    macro->line = cpp->line;

    // Skip macro name
    pos += len;

    // Handle argument list for function-like macros
    if ( *pos == '(' ) {
        macro->fnlike = true;
        argv_size = 2;
        argv = cpp_heap_calloc(&cpp->output_heap, argv_size, sizeof(char*));
        // Skip opening '('
        ++pos;
        // Parse argument list
        for (;;) {
            pos = cpp_skipspace(pos);
            cpp_identify_token(pos, &tag, &len);
            if ( tag == T_PAREN_CLOSE ) {
                break;
            }
            if ( macro->argc ) {
                if ( tag != T_COMMA ) {
                    cpp_msg_new(cpp);
                    cpp_msg_cats(cpp, "Expected ',' or ')' after argument name in macro declaration");
                    if ( tag != T_END ) {
                        cpp_msg_cats(cpp, ", found ");
                        cpp_msg_cats(cpp, cpp_strtag(tag));
                    }
                    pptoken = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
                    cpp_output_append(cpp, pptoken);
                    return;
                }
                pos += len;
                pos = cpp_skipspace(pos);
                cpp_identify_token(pos, &tag, &len);
            }
            if ( tag != T_IDENT && tag != T_ELLIPSIS ) {
                cpp_msg_new(cpp);
                cpp_msg_cats(cpp, "Unexpected pptoken in macro argument declaration: ");
                cpp_msg_cats(cpp, cpp_strtag(tag));
                pptoken = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
                cpp_output_append(cpp, pptoken);
                return;
            }
            if ( macro->varargs ) {
                msg = "'...' can only be last argument";
                pptoken = cpp_token_news_recycle(cpp, T_ERROR, msg);
                cpp_output_append(cpp, pptoken);
                return;
            }
            if ( tag == T_ELLIPSIS ) {
                macro->varargs = true;
            }
            if ( macro->argc == argv_size ) {
                char** v;
                v = cpp_heap_calloc(&cpp->output_heap, argv_size*2, sizeof(char*));
                memcpy(v, argv, argv_size * sizeof(char*));
                argv = v;
                argv_size *= 2;
            }
            argv[macro->argc] = cpp_heap_strndup(&cpp->unit_heap, pos, len);
            ++(macro->argc);
            pos += len;
        }
        // Skip closing ')'
        pos += len;
        // Create vector of exact size
        bytes = macro->argc * sizeof(char*);
        macro->argv = cpp_heap_malloc(&cpp->unit_heap, bytes);
        memcpy(macro->argv, argv, bytes);
    }

    // Get replacement list
    rev_body = NULL;
    space = false;
    pos = cpp_skipspace(pos);
    while ( cpp_identify_token(pos,&tag,&len), tag != T_END ) {
        if ( tag == T_SPACE ) {
            space = true;
        } else {
            if ( space ) {
                mtoken = cpp_heap_calloc(&cpp->unit_heap, 1, sizeof(MACROTOKEN));
                mtoken->tag = T_SPACE;
                mtoken->str = " ";
                mtoken->next = rev_body;
                rev_body = mtoken;
                space = false;
            }
            mtoken = cpp_heap_calloc(&cpp->unit_heap, 1, sizeof(MACROTOKEN));
            mtoken->tag = tag;
            mtoken->str = cpp_heap_strndup(&cpp->unit_heap, pos, len);
            mtoken->next = rev_body;
            rev_body = mtoken;
            if ( tag == T_IDENT ) {
                if ( strcmp(mtoken->str, "__VA_ARGS__") == 0 ) {
                    if ( !macro->varargs ) {
                        cpp_msg_new(cpp);
                        cpp_msg_cats(cpp, "'__VA_ARGS__' only allowed in macros with '...' argument");
                        pptoken = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
                        cpp_output_append(cpp, pptoken);
                        return;
                    }
                    mtoken->tag = T_ARGUMENT;
                    mtoken->argno = macro->argc - 1;
                } else {
                    size_t i;
                    for ( i = 0 ; i < macro->argc ; i++ ) {
                        if ( strcmp(mtoken->str, macro->argv[i]) == 0 ) {
                            mtoken->tag = T_ARGUMENT;
                            mtoken->argno = i;
                            break;
                        }
                    }
                }
            }
            prev_tag = tag;
        }
        pos += len;
    }

    // Process replacement again, validate and reverse
    while ( rev_body ) {
        mtoken = rev_body;
        rev_body = rev_body->next;
        if ( mtoken->tag == T_HASH_HASH ) {
            // Remove spaces around '##'
            if ( rev_body && rev_body->tag == T_SPACE ) {
                rev_body = rev_body->next;
            }
            if ( macro->body && macro->body->tag == T_SPACE ) {
                macro->body = macro->body->next;
            }
            // Validate position
            if ( rev_body == NULL ) {
                msg = "'##' cannot be first in replacement list";
                pptoken = cpp_token_news_recycle(cpp, T_ERROR, msg);
                cpp_output_append(cpp, pptoken);
                return;
            }
            if ( macro->body == NULL ) {
                msg = "'##' cannot be last in replacement list";
                pptoken = cpp_token_news_recycle(cpp, T_ERROR, msg);
                cpp_output_append(cpp, pptoken);
                return;
            }
        }
        if ( mtoken->tag == T_HASH ) {
            // Remove space between '#' and argument
            if ( macro->body
              && macro->body->tag == T_SPACE
              && macro->body->next
              && macro->body->next->tag == T_ARGUMENT
              )
            {
                macro->body = macro->body->next;
            }
            // Combine '#' 'arg' into one stringize token
            if ( macro->body && macro->body->tag == T_ARGUMENT ) {
                char* str;
                mtoken->argno = macro->body->argno;
                mtoken->tag = T_STRINGIZE;
                str = (char*)cpp_heap_malloc(&cpp->unit_heap, strlen(macro->body->str)+2);
                strcpy(str, "#");
                strcat(str, macro->body->str);
                mtoken->str = str;
                macro->body = macro->body->next;
            }
        }
        mtoken->next = macro->body;
        macro->body = mtoken;
    }

    cpp_define_macro(cpp, macro);
}


// Parse '#undef' directive
void cpp_parse_undef(CPP* cpp, const char* pos)
{
    PPTOKEN*    pptoken;
    int         tag;
    size_t      len;
    char*       name;

    // Get name of macro to define
    pos = cpp_skipspace(pos);
    cpp_identify_token(pos, &tag, &len);
    if ( tag != T_IDENT ) {
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "Expected identifier after '#undef'");
        if ( tag != T_END ) {
            cpp_msg_cats(cpp, ", found ");
            cpp_msg_cats(cpp, cpp_strtag(tag));
        }
        pptoken = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
        cpp_output_append(cpp, pptoken);
        return;
    }

    // Remove the macro
    name = cpp_heap_strndup(&cpp->output_heap, pos, len);
    pos += len;
    cpp_undefine_macro(cpp, name);

    // Check for unexpected tokens
    pos = cpp_skipspace(pos);
    if ( *pos ) {
        cpp_identify_token(pos, &tag, &len);
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "Expected token after '#undef ");
        cpp_msg_cats(cpp, name);
        cpp_msg_cats(cpp, "': ");
        cpp_msg_cats(cpp, cpp_strtag(tag));
        pptoken = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
        cpp_output_append(cpp, pptoken);
    }
}


// Create special macro replacement list
PPTOKEN* cpp_special_macro(CPP* cpp, MACRO* macro)
{
    char        buffer[sizeof(cpp->line)*3+2];
    PPTOKEN*    token;
    const char* pos;

    switch ( macro->special ) {
        case BUILTIN_LINE:
            sprintf(buffer, "%d", cpp->line);
            token = cpp_token_news(cpp, T_NUMLIT, buffer);
            break;
        case BUILTIN_FILE:
            cpp_buffer_clear(&cpp->scratch);
            cpp_buffer_catc(&cpp->scratch, '"');
            for ( pos = cpp->file ; *pos ; pos++ ) {
                if ( *pos == '\\' || *pos == '\"' ) {
                    cpp_buffer_catc(&cpp->scratch, '\\');
                }
                cpp_buffer_catc(&cpp->scratch, *pos);
            }
            cpp_buffer_catc(&cpp->scratch, '"');
            token = cpp_token_news(cpp, T_STRLIT, cpp_buffer_gets(&cpp->scratch));
            break;
        default:
            token = NULL;
            assert(0);
            break;
    }

    return token;
}


static void cpp_define_builtin(CPP* cpp, const char* name, int special, int tag, const char* value)
{
    MACRO* macro;

    macro = (MACRO*)cpp_heap_calloc(&cpp->unit_heap, 1, sizeof(MACRO));
    macro->name = name;
    macro->file = "(built-in)";
    macro->special = special;
    macro->body = (MACROTOKEN*)cpp_heap_calloc(&cpp->unit_heap, 1, sizeof(MACROTOKEN));
    macro->body->tag = tag;
    macro->body->str = cpp_heap_strdup(&cpp->unit_heap, value);
    cpp_define_macro(cpp, macro);
}


static void cpp_define_other(CPP* cpp, const char* name, const char* value)
{
    MACRO*      macro;
    MACROTOKEN* token;
    MACROTOKEN* last = NULL;
    int         tag;
    size_t      len;

    macro = (MACRO*)cpp_heap_calloc(&cpp->unit_heap, 1, sizeof(MACRO));
    macro->name = name;
    macro->file = "(config)";

    while ( cpp_identify_token(value,&tag,&len), tag != T_END ) {
        token = (MACROTOKEN*)cpp_heap_calloc(&cpp->unit_heap, 1, sizeof(MACROTOKEN));
        token->tag = tag;
        if ( tag == T_SPACE ) {
            token->str = " ";
        } else {
            token->str = cpp_heap_strndup(&cpp->unit_heap, value, len);
        }
        value += len;
        if ( macro->body ) {
            last->next = token;
        } else {
            macro->body = token;
        }
        last = token;
    }

    cpp_define_macro(cpp, macro);
}


// Prepare for new translation unit
void cpp_macro_prepare(CPP* cpp)
{
    time_t      now;
    struct tm*  date_time;
    char        buf[20];
    DEFINE*     define;

    // Define __DATE__ and __TIME__ are regular macros
    now = time(NULL);
    date_time = localtime(&now);
    if ( date_time == NULL ) {
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "localtime() failed");
        cpp_fatal(cpp, cpp_msg_gets(cpp));
    }
    strftime(buf, sizeof(buf), "\"%b %d %Y\"", date_time);
    if ( buf[5] == '0' ) {
        buf[5] = ' '; // mandated by standard
    }
    cpp_define_builtin(cpp, "__DATE__", NOT_SPECIAL, T_STRLIT, buf);
    strftime(buf, sizeof(buf), "\"%H:%M:%S\"", date_time);
    cpp_define_builtin(cpp, "__TIME__", NOT_SPECIAL, T_STRLIT, buf);

    // Define __FILE__ and __LINE__ specially
    cpp_define_builtin(cpp, "__FILE__", BUILTIN_FILE, T_STRLIT, "");
    cpp_define_builtin(cpp, "__LINE__", BUILTIN_LINE, T_NUMLIT, "0");

    // Definitions from configuration
    for ( define = cpp->defines ; define ; define = define->next ) {
        cpp_undefine_macro(cpp, define->name);
        if ( define->value ) {
            cpp_define_other(cpp, define->name, define->value);
        }
    }
}
