/******************************************************************************

    $Id: main.c,v 1.4 2003/08/28 19:12:27 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/main.c,v $

******************************************************************************/

#include "cpp.h"

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

enum {
    PREPROCESS,
    TOKENIZE,
    MACRO_DUMP,
    STATISTICS
};

int main(int argc, char* argv[])
{
    int         exitcode = 0;
    const char* msg;
    CPP*        cpp;
    FILE*       err = stderr;
    const char* path;
    FILE*       in;
    int         nfiles = 0;
    TOKEN       token;
    int         mode = PREPROCESS;
    int         verbose = 0;
    int         config_vc7 = 0;
    
    // Initialize preprocessor
    cpp = cpp_new();
    if ( cpp == NULL ) {
        fprintf(stderr, "Out of memory");
        return 1;
    }

    // Get rid of unpredictable argv[0]
    if ( argc > 0 ) {
        --argc;
        ++argv;
    } else {
        argc = 0;
        argv = NULL;
    }

    // Parse options
    while ( argc && *argv[0] == '-' ) {
        const char* opt = argv[0];
        --argc;
        ++argv;
        if ( strcmp(opt, "--") == 0 ) {
            // POSIX end-of-options
            break;
        } else if ( strcmp(opt, "-m") == 0 ) {
            mode = MACRO_DUMP;
        } else if ( strcmp(opt, "-t") == 0 ) {
            mode = TOKENIZE;
        } else if ( strcmp(opt, "-s") == 0 ) {
            mode = STATISTICS;
        } else if ( strcmp(opt, "-v") == 0 ) {
            verbose = 1;
        } else if ( strcmp(opt, "-vc7") == 0 ) {
            config_vc7 = 1;
        } else if ( opt[1] && strchr("DIU", opt[1]) ) {
            switch ( opt[1] ) {
                case 'D':
                    msg = cpp_define(cpp, opt+2);
                    break;
                case 'I':
                    msg = cpp_include(cpp, opt+2);
                    break;
                case 'U':
                    msg = cpp_undefine(cpp, opt+2);
                    break;
                default:
                    msg = "Internal error";
                    break;
            }
            if ( msg ) {
                fprintf(err, "%s: %s\n", msg, opt);
                exitcode = 1;
                break;
            }
        } else {
            fprintf(err, "Unrecognized option: %s\n", opt);
            exitcode = 1;
            break;
        }
    }

    // VC7 like configuration
    if ( config_vc7 ) {
        cpp_include_list(cpp, getenv("INCLUDE"));
        cpp_define(cpp, "_M_IX86=600");
        cpp_define(cpp, "_MFC_VER=0x0700");
        cpp_define(cpp, "_WIN32=");
    }

    // Preprocess files
    while ( exitcode==0 && (nfiles==0 || argc) ) {
        ++nfiles;
        // Open file
        if ( argc == 0 ) {
            path = "(stdin)";
            in = stdin;
        } else {
            path = argv[0];
            in = NULL;
            --argc;
            ++argv;
        }
        msg = cpp_open(cpp, path, in);
        if ( msg ) {
            fprintf(stderr, "Error opening \"%s\": %s\n", path, msg);
            exitcode = 1;
            break;
        }
        // Preprocess file
        while ( (msg=cpp_gettoken(cpp, &token))==NULL && token.tag != T_END ) {
            switch ( mode ) {
                case TOKENIZE:
                    if ( token.tag != T_SPACE ) {
                        printf("%s(%d) %-14s %s\n", token.file, token.line, cpp_strtag(token.tag), token.str);
                    } else if ( verbose ) {
                        const char* pos = token.str;
                        printf("%s(%d) %-14s \"", token.file, token.line, cpp_strtag(token.tag));
                        while ( *pos ) {
                            if ( *pos == '\n' ) {
                                printf("\\n");
                            } else {
                                putchar(*pos);
                            }
                            ++pos;
                        }
                        printf("\"\n");
                    }
                    break;
                case PREPROCESS:
                    switch ( token.tag ) {
                        case T_ERROR:
                            printf("%s(%d) %s\n", token.file, token.line, token.str);
                            break;
                        case T_PRAGMA:
                            printf("#pragma %s\n", token.str);
                            break;
                        default:
                            fputs(token.str, stdout);
                            break;
                    }
                    break;
                case MACRO_DUMP:
                    if ( token.tag == T_ERROR ) {
                        printf("%s(%d) %s\n", token.file, token.line, token.str);
                    }
                    break;
                case STATISTICS:
                    break;
            }
        }
        if ( msg ) {
            fprintf(stderr, "%s\n", msg);
            exitcode = 1;
        }
        switch ( mode ) {
            case MACRO_DUMP:
                cpp_dump_macros(cpp, stdout, verbose);
                break;
            case STATISTICS:
                cpp_dump_memstat(cpp, stdout);
                break;
        }
    }

    // Clean up and return
    cpp_delete(cpp);
    return exitcode;
}
