/******************************************************************************

    $Id: parse.c,v 1.4 2003/08/28 19:12:27 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/parse.c,v $

******************************************************************************/

#include "priv.h"


// Directive numbers
enum {
    DIR_NONE,
    DIR_DEFINE,
    DIR_ELIF,
    DIR_ELSE,
    DIR_ENDIF,
    DIR_ERROR,
    DIR_IF,
    DIR_IFDEF,
    DIR_IFNDEF,
    DIR_INCLUDE,
    DIR_LINE,
    DIR_PRAGMA,
    DIR_UNDEF
};


// Map identifier to directive number
static int cpp_directive(const char* pos, size_t len)
{
    #define HASH(len,first)     ((len) << 8)|(unsigned char)(first)
    #define MAP(ident,len,num)  if (memcmp(ident,pos,len)==0) return num

    switch ( HASH(len,*pos) ) {
        case HASH(6,'d'):
            MAP("define", 6, DIR_DEFINE);
            break;
        case HASH(4,'e'):
            MAP("elif", 4, DIR_ELIF);
            MAP("else", 4, DIR_ELSE);
            break;
        case HASH(5,'e'):
            MAP("error", 5, DIR_ERROR);
            MAP("endif", 5, DIR_ENDIF);
            break;
        case HASH(2,'i'):
            MAP("if", 2, DIR_IF);
            break;
        case HASH(5,'i'):
            MAP("ifdef", 5, DIR_IFDEF);
            break;
        case HASH(6,'i'):
            MAP("ifndef", 6, DIR_IFNDEF);
            break;
        case HASH(7,'i'):
            MAP("include", 7, DIR_INCLUDE);
            break;
        case HASH(4,'l'):
            MAP("line", 4, DIR_LINE);
            break;
        case HASH(6,'p'):
            MAP("pragma", 6, DIR_PRAGMA);
            break;
        case HASH(5,'u'):
            MAP("undef", 5, DIR_UNDEF);
            break;
    }
    return DIR_NONE;
}


static void cpp_parse_directive(CPP* cpp)
{
    PPTOKEN*    token;
    const char* pos;
    int         tag;
    size_t      len;

    // Replace directive with newlines in output
    cpp->newlines = cpp_token_newnl(cpp, cpp_nlcount(cpp->input_ptr));
    pos = cpp->input_ptr;
    cpp->input_ptr = "";

    // Get identifier with directive name
    pos = cpp_skipspace(pos);
    cpp_identify_token(pos, &tag, &len);
    if ( tag == T_END ) {
        // Empty directive
        cpp_output_append(cpp, cpp->newlines);
        return;
    }
    if ( tag != T_IDENT ) {
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "Expected identifier after '#', found ");
        cpp_msg_cats(cpp, cpp_strtag(tag));
        token = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
        cpp_output_append(cpp, cpp->newlines);
        cpp_output_append(cpp, token);
        return;
    }

    // Individual handling of directives
    switch ( cpp_directive(pos,len) ) {
        case DIR_DEFINE:
            if ( cpp->is_including ) {
                cpp_parse_define(cpp, pos+len);
            }
            break;
        case DIR_UNDEF:
            if ( cpp->is_including ) {
                cpp_parse_undef(cpp, pos+len);
            }
            break;
        case DIR_PRAGMA:
            if ( cpp->is_including ) {
                cpp_parse_pragma(cpp, pos+len);
            }
            break;
        case DIR_LINE:
            if ( cpp->is_including ) {
                cpp_parse_line(cpp, pos+len);
            }
            break;
        case DIR_INCLUDE:
            if ( cpp->is_including ) {
                cpp_parse_include(cpp, pos+len);
            }
            break;
        case DIR_IF:
            cpp_parse_if(cpp, pos+len);
            cpp->is_including = cpp_is_conditional_true(cpp);
            break;
        case DIR_IFDEF:
            cpp_parse_ifdef(cpp, pos+len);
            cpp->is_including = cpp_is_conditional_true(cpp);
            break;
        case DIR_IFNDEF:
            cpp_parse_ifndef(cpp, pos+len);
            cpp->is_including = cpp_is_conditional_true(cpp);
            break;
        case DIR_ELIF:
            cpp_parse_elif(cpp, pos+len);
            cpp->is_including = cpp_is_conditional_true(cpp);
            break;
        case DIR_ELSE:
            cpp_parse_else(cpp, pos+len);
            cpp->is_including = cpp_is_conditional_true(cpp);
            break;
        case DIR_ENDIF:
            cpp_parse_endif(cpp, pos+len);
            cpp->is_including = cpp_is_conditional_true(cpp);
            break;
        default:
            if ( cpp->is_including ) {
                cpp_msg_new(cpp);
                cpp_msg_cats(cpp, "Unrecognized directive: '#");
                cpp_msg_catn(cpp, pos, len);
                cpp_msg_cats(cpp, "'");
                token = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
                cpp_output_append(cpp, token);
            }
            break;
    }

    cpp_output_append(cpp, cpp->newlines);
}


void cpp_parse_identifier(CPP* cpp, PPTOKEN* ident)
{
    MACRO*      macro;
    PPTOKEN*    token;
    PPTOKEN*    expanded;
    PPLIST      list;
    int         paren;
    size_t      nlcount = 0;

    macro = cpp_find_macro(cpp, ident->token.str);
    if ( macro == NULL ) {
        cpp_output_append(cpp, ident);
        return;
    }

    if ( macro->fnlike ) {

        // Check for '(' following the macro name
        cpp_list_reset(&list);
        cpp_list_append(&list, ident);
        while ( token=cpp_input_entok(cpp), token->token.tag == T_SPACE ) {
            nlcount += cpp_nlcount(token->token.str);
            cpp_list_append(&list, token);
        }
        if ( token->token.tag != T_PAREN_OPEN ) {
            // It wasn't a macro invocation after all, emit the
            // identifier and all spaces that followed it.
            cpp->push_back = token;
            while ( (token = cpp_list_getfirst(&list)) != NULL ) {
                cpp_output_append(cpp, token);
            }
            return;
        }

        // Make list look like: identifier -> '(' -> NULL
        list.first->next = token;
        list.last = token;

        // Collect arguments
        paren = 1;
        do {
            token = cpp_input_entok(cpp);
            switch ( token->token.tag ) {
                case T_SPACE:
                    nlcount += cpp_nlcount(token->token.str);
                    token->token.str = " ";
                    break;
                case T_PAREN_OPEN:
                    ++paren;
                    break;
                case T_PAREN_CLOSE:
                    --paren;
                    break;
                case T_END:
                    cpp_msg_new(cpp);
                    cpp_msg_cats(cpp, "Argument list to macro '");
                    cpp_msg_cats(cpp, ident->token.str);
                    cpp_msg_cats(cpp, "' never ended with ')'");
                    token = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
                    token->token.file = ident->token.file;
                    token->token.line = ident->token.line;
                    cpp_output_append(cpp, token);
                    return;
            }
            cpp_list_append(&list, token);
        } while ( paren );

        // Expand
        expanded = cpp_expand_list(cpp, list.first, NULL);
    } else {
        expanded = cpp_expand_macro(cpp, macro, NULL, NULL);
    }

    cpp_output_appendlist(cpp, expanded);

    // White space after expanded macro
    if ( nlcount ) {
        token = cpp_token_newnl(cpp, nlcount);
        cpp_output_append(cpp, token);
    }
}


static void cpp_checked_pop(CPP* cpp)
{
    CONDITIONAL*    cond;
    PPTOKEN*        token;
    INPUT*          input   = cpp->input;
    size_t          ofs     = 0;

    if ( input->parent ) {
        ofs = input->parent->cond_top;
        while ( input->cond_top > ofs ) {
            cond = cpp->cond_stack + input->cond_top - 1;
            cpp_msg_new(cpp);
            cpp_msg_cats(cpp, "'#");
            cpp_msg_cats(cpp, cond->ident);
            cpp_msg_cats(cpp, "' at ");
            cpp_msg_cats(cpp, "has no '#endif'");
            token = cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp));
            cpp_output_append(cpp, token);
            --(input->cond_top);
        }
    }
    cpp_input_pop(cpp);
}


void cpp_parse(CPP* cpp)
{
    PPTOKEN* token;

    // Get next token
    for (;;) {
        if ( cpp->input == NULL ) {
            token = cpp_token_new(cpp, T_END);
            break;
        }
        token = cpp_input_entok(cpp);
        if ( token->token.tag != T_END ) {
            break;
        }
        cpp_checked_pop(cpp);
    }

    switch ( token->token.tag ) {
        case T_BOL_HASH:
            cpp_parse_directive(cpp);
            break;
        case T_IDENT:
            if ( cpp->is_including ) {
                cpp_parse_identifier(cpp, token);
            }
            break;
        case T_SPACE:
        case T_END:
            cpp_output_append(cpp, token);
            break;
        default:
            if ( cpp->is_including ) {
                cpp_output_append(cpp, token);
            }
            break;
    }
}
