/******************************************************************************

    $Id: priv.h,v 1.9 2003/08/28 21:59:29 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/priv.h,v $

    Internals of the CPP library
    
******************************************************************************/

/******************************************************************************
    Includes
******************************************************************************/

#include "cpp.h"

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <stddef.h>
#include <setjmp.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

/******************************************************************************
    Constants
******************************************************************************/

// Size of memory blocks to allocated for HEAPs.
#define SYSMEM_BLOCKSIZE 4096

// C++ style bool
#define bool    int
#define false   0
#define true    1

// Internal token tags
enum {
    T_BOL_HASH = T_BAD + 1,     // '#' at beginning of line
    T_STRINGIZE,                // Stringize macro argument
    T_ARGUMENT,                 // Substitute with argument
    T_PLACEMARKER,              // Place marker in fnlike macro expansion
    T_FINAL_IDENT               // Not available for substituion
};

/******************************************************************************
    Character tests
******************************************************************************/

#define ISSPACE(ch)     (isspace((unsigned char)(ch)))
#define ISALPHA(ch)     (isalpha((unsigned char)(ch)))
#define ISDIGIT(ch)     (isdigit((unsigned char)(ch)))
#define ISODIGIT(ch)    (isdigit((unsigned char)(ch)) && (ch)!='8' && (ch)!='9')
#define ISXDIGIT(ch)    (isxdigit((unsigned char)(ch)))
#define ISIDENT1(ch)    (isalpha((unsigned char)(ch)) || (ch)=='_')
#define ISIDENT(ch)     (isalnum((unsigned char)(ch)) || (ch)=='_')

/******************************************************************************
    Forwards
******************************************************************************/

typedef struct tagBUFFER        BUFFER;
typedef struct tagCONDITIONAL   CONDITIONAL;
typedef struct tagDEFINE        DEFINE;
typedef struct tagEXPR          EXPR;
typedef struct tagFRAME         FRAME;
typedef struct tagHEAP          HEAP;
typedef struct tagINPUT         INPUT;
typedef struct tagINCDIR        INCDIR;
typedef struct tagMACRO         MACRO;
typedef struct tagMACROTOKEN    MACROTOKEN;
typedef struct tagPPTOKEN       PPTOKEN;
typedef struct tagPPLIST        PPLIST;
typedef struct tagSYSMEM        SYSMEM;
typedef struct tagVALUE         VALUE;

/******************************************************************************
    Memory Management - data structures
******************************************************************************/

// Buffer for dynamic strings
struct tagBUFFER {
    CPP*            cpp;                // Owning preprocessor
    size_t          size;               // Allocated size
    size_t          len;                // Current length
    char*           data;               // Memory allocated using malloc() or realloc()
};

// Heap to allocate from
struct tagHEAP {
    CPP*            cpp;                // Owning preprocessor
    SYSMEM*         top;                // Stack of blocks allocated using malloc()
    size_t          size;               // Current size
    size_t          top_size;           // Max size encountered until now
};

/******************************************************************************
    Configuration - data structures
******************************************************************************/

// Include directory
struct tagINCDIR {
    INCDIR*         next;               // -> next in search path
    const char*     path;               // Path with trailing directory separator
};

// '-D' or '-U' option
struct tagDEFINE {
    DEFINE*         next;               // Next in list
    const char*     name;               // Name of macro
    const char*     value;              // Value to assign, NULL if undef
};

/******************************************************************************
    Preprocessing - data structures
******************************************************************************/

// Input file
struct tagINPUT {
    INPUT*          parent;             // File included from
    FILE*           fp;                 // Input stream
    bool            owner;              // Need to fclose(fp)?
    const char*     path;               // Name of file opened
    const char*     file;               // Current value of __FILE__
    int             nlcount;            // Number of newlines seen
    size_t          cond_top;           // Top of condition stack
};

// Internal preprocessing token
struct tagPPTOKEN {
    PPTOKEN*        next;               // -> next in list
    TOKEN           token;              // Token to return
};

// Token in macro replacement list
struct tagMACROTOKEN {
    MACROTOKEN*     next;               // -> next in list
    int             tag;                // Tag of token
    const char*     str;                // Text from source file
    size_t          argno;              // If 'tag' is T_ARGUMENT or T_STRINGIZE
};

// List of preprocessing tokens
struct tagPPLIST {
    PPTOKEN*        first;              // -> first in list
    PPTOKEN*        last;               // -> last in list
};

// Macro definition
struct tagMACRO {
    MACRO*          next;               // -> next in hash chain
    const char*     name;               // Name of macro
    const char*     file;               // __FILE__ where defined
    int             line;               // __LINE__ where defined
    int             special;            // Number of special macro
    bool            fnlike;             // Function-like?
    bool            varargs;            // Is last argument '...'?
    size_t          argc;               // Number of arguments
    char**          argv;               // Vector of argument names
    size_t          expand_count;       // Number of expansions made
    MACROTOKEN*     body;               // Replacement list
};

// Frame in macro expansion 
struct tagFRAME {
    FRAME*          parent;             // -> parent frame
    MACRO*          macro;              // Macro being expanded
    const char*     name;               // Name of macro (saves one indirection)
};

// Conditional block
struct tagCONDITIONAL {
    const char*     file;               // __FILE__ where introduced
    int             line;               // __LINE__ where introduced
    const char*     ident;              // "if", "ifdef", or "ifndef"
    bool            skip_rest;          // Rest of block should be ignored
    bool            is_true;            // Is true right now?
    const char*     else_file;          // __FILE__ where '#else' was seen
    bool            else_line;          // __LINE__ where '#else' was seen
};


/******************************************************************************
    Main preprocessor context
******************************************************************************/

struct tagCPP {
    // Error handling
    jmp_buf         fatal_exit;         // Where to jump in case of fatal error
    const char*     fatal_msg;          // Error message to return
    // Memory management
    size_t          blocks_allocated;   // Blocks allocated using malloc() and friends
    size_t          alloc_count;        // Number of allocations
    size_t          dealloc_count;      // Number of deallocations
    SYSMEM*         sysmem_freelist;    // Free list for SYSMEM structures.
    INPUT*          input_freelist;     // Free list for INPUT structures.
    HEAP            cpp_heap;           // Allocations per CPP instance.
    HEAP            unit_heap;          // Allocations per translation unit.
    HEAP            output_heap;        // Output tokens - often cleared
    // Configuration
    INCDIR*         includes;           // List of include directories
    DEFINE*         defines;            // List of '-D' and '-U' options
    bool            expand_pragma;      // Expand #pragma? (except STDC)
    // Parsing
    BUFFER          input_buf;          // Input buffer
    const char*     input_ptr;          // Current position within 'input_buf'
    INPUT*          input;              // Stack of input files
    PPTOKEN*        push_back;          // Pushed back token
    PPTOKEN*        newlines;           // Newlines to be emitted after directive
    int             output_state;       // State of output queue
    PPLIST          output_queue;       // Queue of tokens to output
    BUFFER          scratch;            // Temporary for various operations
    BUFFER          msg_buf;            // For building error messages
    const char*     file;               // Current value of __FILE__
    int             line;               // Current value of __LINE__
    size_t          macro_count;        // Number of macros
    size_t          macro_hsize;        // Size of hash table with macros (#buckets)
    MACRO**         macro_htab;         // Hash table with macros
    CONDITIONAL*    cond_stack;         // Stack of conditionals
    size_t          cond_stack_size;    // Allocated size of 'cond_stack'
    bool            is_including;       // Is conditional true right now?
};


/******************************************************************************
    Prototypes    
******************************************************************************/

// alloc.c
void* cpp_malloc(CPP* cpp, size_t size);
void* cpp_calloc(CPP* cpp, size_t num, size_t size);
void* cpp_realloc(CPP* cpp, void* ptr, size_t size);
void cpp_free(CPP* cpp, void* ptr);
void cpp_out_of_memory(CPP* cpp);
void* cpp_safe_malloc(CPP* cpp, size_t size);
void* cpp_safe_calloc(CPP* cpp, size_t num, size_t size);
void* cpp_safe_realloc(CPP* cpp, void* ptr, size_t size);

// buffer.c
void cpp_buffer_init(CPP* cpp, BUFFER* buf);
void cpp_buffer_cleanup(BUFFER* buf);
void cpp_buffer_grow(BUFFER* buf, size_t need);
void cpp_buffer_clear(BUFFER* buf);
void cpp_buffer_catc(BUFFER* buf, char ch);
void cpp_buffer_catn(BUFFER* buf, const char* str, size_t len);
void cpp_buffer_cats(BUFFER* buf, const char* str);
void cpp_buffer_cati(BUFFER* buf, int number);
void cpp_buffer_insertc(BUFFER* buf, size_t pos, char ch);
void cpp_buffer_removec(BUFFER* buf, size_t pos);
char* cpp_buffer_gets(BUFFER* buf);

// cond.c
bool cpp_is_conditional_true(CPP* cpp);
void cpp_parse_if(CPP* cpp, const char* pos);
void cpp_parse_ifdef(CPP* cpp, const char* pos);
void cpp_parse_ifndef(CPP* cpp, const char* pos);
void cpp_parse_elif(CPP* cpp, const char* pos);
void cpp_parse_else(CPP* cpp, const char* pos);
void cpp_parse_endif(CPP* cpp, const char* pos);

// config.c
void cpp_config_define(CPP* cpp, const char* name, size_t len, const char* value);
void cpp_config_include(CPP* cpp, const char* path);
void cpp_config_include_list(CPP* cpp, const char* list);

// cpp.c
void cpp_fatal(CPP* cpp, const char* msg);

// dump.c
void cpp_dump_expression(int level, EXPR* expr);

// entok.c
size_t cpp_header_name_len(const char* pos);
size_t cpp_line_name_len(const char* pos);
size_t cpp_digit_sequence_len(const char* pos);
size_t cpp_universal_char_len(const char* pos);
size_t cpp_identifier_len(const char* pos);
size_t cpp_text_constant_len(const char* pos);
size_t cpp_pp_number_len(const char* pos);
void cpp_identify_token(const char* pos, int* tag, size_t* len);

// expand.c
void cpp_list_flatten(BUFFER* dest, PPTOKEN* list);
PPTOKEN* cpp_expand_list(CPP* cpp, PPTOKEN* list, FRAME* parent);
PPTOKEN* cpp_expand_macro(CPP* cpp, MACRO* macro, PPTOKEN** argv, FRAME* parent);

// expr.c
bool cpp_eval_if_expression(CPP* cpp, const char* ident, const char* pos);

// heap.c
void cpp_heap_init(CPP* cpp, HEAP* heap);
void cpp_heap_cleanup(CPP* cpp);
void cpp_heap_freetop(HEAP* heap);
void cpp_heap_clear(HEAP* heap);
void* cpp_heap_malloc(HEAP* heap, size_t size);
void* cpp_heap_calloc(HEAP* heap, size_t num, size_t size);
void* cpp_heap_mark(HEAP* heap);
void cpp_heap_release(HEAP* heap, void* mark);
void* cpp_heap_strndup(HEAP* heap, const char* str, size_t len);
void* cpp_heap_strdup(HEAP* heap, const char* str);

// include.c
void cpp_parse_include(CPP* cpp, const char* pos);

// input.c
void cpp_input_push(CPP* cpp, const char* path);
void cpp_input_pop(CPP* cpp);
void cpp_input_popall(CPP* cpp);
PPTOKEN* cpp_input_entok(CPP* cpp);

// line.c
void cpp_parse_line(CPP* cpp, const char* pos);

// macro.c
bool cpp_is_macro_equal(MACRO* macro1, MACRO* macro2);
bool cpp_is_macro_reserved(const char* name);
MACRO* cpp_find_macro(CPP* cpp, const char* name);
void cpp_undefine_macro(CPP* cpp, const char* name);
void cpp_define_macro(CPP* cpp, MACRO* macro);
void cpp_parse_define(CPP* cpp, const char* pos);
void cpp_parse_undef(CPP* cpp, const char* pos);
void cpp_macro_prepare(CPP* cpp);
PPTOKEN* cpp_special_macro(CPP* cpp, MACRO* macro);

// msg.c
void cpp_msg_new(CPP* cpp);
void cpp_msg_newfl(CPP* cpp, const char* file, int line);
void cpp_msg_catc(CPP* cpp, char ch);
void cpp_msg_catn(CPP* cpp, const char* str, size_t len);
void cpp_msg_cats(CPP* cpp, const char* str);
void cpp_msg_cati(CPP* cpp, int number);
void cpp_msg_catfl(CPP* cpp, const char* file, int line);
const char* cpp_msg_gets(CPP* cpp);

// output.c
void cpp_output_append(CPP* cpp, PPTOKEN* token);
void cpp_output_appendlist(CPP* cpp, PPTOKEN* tokens);
PPTOKEN* cpp_output_get(CPP* cpp);

// parse.c
void cpp_parse(CPP* cpp);

// pragma.c
void cpp_parse_pragma(CPP* cpp, const char* pos);
void cpp_extract_pragma_operator(CPP* cpp, PPLIST* list);

// listfix.c
void cpp_listfix(CPP* cpp, PPLIST* list);

// path.c
int cpp_is_dirsep(char ch);
char cpp_dirsep(void);
char cpp_listsep(void);
int cpp_pathcmp(const char* path1, const char* path2);
bool cpp_is_path_abs(const char* path);
const char* cpp_basename(const char* path);
bool cpp_exists(const char* path);

// str.c
size_t cpp_nlcount(const char* str);
const char* cpp_skipspace(const char* str);
size_t cpp_hash(const char* str);
size_t cpp_hashn(const char* str, size_t len);

// token.c
PPTOKEN* cpp_token_new(CPP* cpp, int tag);
PPTOKEN* cpp_token_news(CPP* cpp, int tag, const char* str);
PPTOKEN* cpp_token_newn(CPP* cpp, int tag, const char* str, size_t len);
PPTOKEN* cpp_token_news_recycle(CPP* cpp, int tag, const char* str);
PPTOKEN* cpp_token_newnl(CPP* cpp, size_t count);
PPTOKEN* cpp_token_clone(CPP* cpp, PPTOKEN* token);
PPTOKEN* cpp_skip_t_space(PPTOKEN* list);
PPTOKEN* cpp_reverse_tokens(PPTOKEN* tokens);
PPTOKEN* cpp_tokenize(CPP* cpp, const char* pos);
void cpp_list_reset(PPLIST* list);
void cpp_list_append(PPLIST* list, PPTOKEN* token);
void cpp_list_appendlist(PPLIST* list, PPTOKEN* tokens);
PPTOKEN* cpp_list_getfirst(PPLIST* list);
